from python_helper import ObjectHelper
from python_framework import Validator, ValidatorMethod, GlobalException, HttpStatus

import SubscriptionDto
import SubscriptionModel

@Validator()
class SubscriptionModelValidator:

    @ValidatorMethod(requestClass=[SubscriptionDto.SubscriptionRequestDto])
    def validateRequestDto(self, dto):
        if ObjectHelper.isNone(dto) or ObjectHelper.isNone(dto.key):
            raise GlobalException(
                message = f'Subscription key cannot be None',
                logMessage = f'Subscription: {dto}',
                status = HttpStatus.BAD_REQUEST
            )


    @ValidatorMethod(requestClass=[SubscriptionDto.SubscriptionRequestDto])
    def validateDoesNotExists(self, dto):
        self.validateRequestDto(dto)
        if self.service.queueModel.existsByKey(dto.key):
            raise GlobalException(
                message = 'Subscription aleady exists',
                logMessage = f'Subscription {dto.key} aleady exists',
                status = HttpStatus.BAD_REQUEST
            )


    @ValidatorMethod(requestClass=[[SubscriptionModel.SubscriptionModel], str])
    def validateIsAtLeastOne(self, modelList, queueKey):
        if ObjectHelper.isEmpty(modelList):
            raise GlobalException(
                message = f'There are no subscriptions listenning to this queue: {queueKey}. Subscriptions: {modelList}',
                status = HttpStatus.BAD_REQUEST
            )


    @ValidatorMethod(requestClass=[[SubscriptionModel.SubscriptionModel]])
    def validateIsExactlyOne(self, modelList):
        if ObjectHelper.isEmpty(modelList):
            raise GlobalException(
                message = 'Subscription not found',
                logMessage = f'An empty list of subscriptions was given: {modelList}',
                status = HttpStatus.BAD_REQUEST
            )
        if 1 < len(modelList):
            raise GlobalException(
                logMessage = f'There are more than one subscription with the same key: {modelList}',
                status = HttpStatus.INTERNAL_SERVER_ERROR
            )



from python_helper import log
from python_framework import HttpStatus, FlaskUtil

from annotation.MessageListenerAnnotation import MessageListener
from annotation.MessageListenerAnnotation import MessageListenerMethod

from enumeration.AccessDomain import AccessDomain
from config import MessageConfig
import MessageDto
import Message


@MessageListener(
    timeout = MessageConfig.LISTENER_TIMEOUT
    , logRequest = True
    , logResponse = True
    , muteLogs = False
)
class MessageListener:

    @MessageListenerMethod(url = '/test/listener/message',
        requestClass=[MessageDto.MessageRequestDto],
        responseClass=[MessageDto.MessageCreationRequestDto],
        apiKeyRequired=[AccessDomain.API]
        , logRequest = True
        , logResponse = True
    )
    def accept(self, dto):
        return self.service.message.globals.api.resource.emitter.message.send(dto), HttpStatus.ACCEPTED


    @MessageListenerMethod(url = '/test/listener/another-message',
        requestClass=[MessageDto.MessageRequestDto],
        responseClass=[MessageDto.MessageCreationRequestDto],
        apiKeyRequired=[AccessDomain.API]
        , logRequest = True
        , logResponse = True
    )
    def anotherAccept(self, dto):
        return self.service.message.globals.api.resource.emitter.message.send(dto), HttpStatus.ACCEPTED



from queue_manager_api.api.src.annotation import MessageEmitterAnnotation
from queue_manager_api.api.src.annotation import MessageListenerAnnotation

from queue_manager_api.api.src.constant import EmitterConstant
from queue_manager_api.api.src.constant import HttpEmitterConstant
from queue_manager_api.api.src.constant import SubscriptionConstant
from queue_manager_api.api.src.constant import MessageConstant

from queue_manager_api.api.src.dto import QueueDto
from queue_manager_api.api.src.dto import SubscriptionDto
from queue_manager_api.api.src.dto import MessageDto

from queue_manager_api.api.src.util import AnnotationUtil
from queue_manager_api.api.src.util import ThreadUtil

from queue_manager_api.api.src.manager.QueueManager import QueueManager

from queue_manager_api.api.src.annotation.MessageEmitterAnnotation import *
from queue_manager_api.api.src.annotation.MessageListenerAnnotation import *



from python_framework import Controller, ControllerMethod, HttpStatus

from enumeration.AccessDomain import AccessDomain


@Controller(url = '/memory', tag='Memory', description='Memory controller')
class MemoryController:

    @ControllerMethod(url = '/',
        apiKeyRequired = [AccessDomain.ADMIN]
        # requestClass = [MessageDto.MessageRequestDto]
        # responseClass = [MessageDto.MessageResponseDto]
        # responseClass = [MessageDto.MessageRequestDto]
        # , logRequest = True
        # , logResponse = True
    )
    def get(self):
        return self.service.memory.findAll(), HttpStatus.OK



from python_framework import SqlAlchemyProxy as sap

from ModelAssociation import QUEUE, SUBSCRIPTION, MODEL
from util import ModelUtil


class QueueModel(MODEL):
    __tablename__ = QUEUE

    id = sap.Column(sap.Integer(), sap.Sequence(f'{__tablename__}{sap.ID}{sap.SEQ}'), primary_key=True)
    key = sap.Column(sap.String(sap.MEDIUM_STRING_SIZE), nullable=False, unique=True)
    originKey = sap.Column(sap.String(sap.MEDIUM_STRING_SIZE), nullable=False)

    subscriptionList = sap.getOneToMany(QUEUE, SUBSCRIPTION, MODEL)

    def __init__(self,
        id = None,
        key = None,
        originKey = None,
        subscriptionList = None
    ):
        self.id = id
        self.key = key
        self.originKey = originKey
        self.subscriptionList = ModelUtil.getOneToManyData(subscriptionList)


    def __repr__(self):
        return f'{self.__tablename__}(id={self.id}, key={self.key}, originKey={self.originKey}, subscriptionKeyList={[subscription.key for subscription in self.subscriptionList]})'



import time

from python_helper import log, ReflectionHelper

try:
    import ThreadUtil
except:
    from queue_manager_api.api.src.util import ThreadUtil


DEFAULT_TIMEOUT = 20


def handleNotRunningThreadDictionary(threadDictionary, threadTimeout=DEFAULT_TIMEOUT):
    while not threadDictionary.isRunning() and not threadDictionary.shouldStopRunning():
        time.sleep(0.01)
    while threadDictionary.isRunning() and not threadDictionary.shouldStopRunning():
        for k in [*threadDictionary.keys()]:
            threadDictionary.get(k).runItIfItsNotRunningYet(threadTimeout=threadTimeout)
            if not threadDictionary.get(k).isAlive():
                threadDictionary.pop(k).kill()
                log.debug(handleNotRunningThreadDictionary, f'The {k}th tread is finished')
        time.sleep(0.01)
    for k in [*threadDictionary.keys()]:
        threadDictionary.pop(k).kill()
    log.debug(handleNotRunningThreadDictionary, f'Thread dictionary finished running')


class ThreadDictionary(dict):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.running = False
        self.shouldStop = False
        log.debug(self.__init__, f'{ReflectionHelper.getName(ThreadDictionary)} created')


    def run(self):
        self.running = True
        self.shouldStop = False


    def isRunning(self):
        return self.running and not self.shouldStop


    def shouldStopRunning(self):
        return True and self.shouldStop


    def kill(self):
        self.shouldStop = True
        self.running = False




class QueueManager:

    def __init__(self, threadTimeout=DEFAULT_TIMEOUT):
        self.threadDictionary = ThreadDictionary()
        self.timeout = threadTimeout
        self.threadDictionaryHandler = ThreadUtil.ApplicationThread(
            handleNotRunningThreadDictionary,
            self.threadDictionary,
            threadTimeout = self.timeout
        )
        log.debug(self.__init__, f'{ReflectionHelper.getName(QueueManager)} created')


    def new(self, target, *args, **kwargs):
        return ThreadUtil.ApplicationThread(target, *args, **kwargs)


    def runInAThread(self, target, *args, **kwargs):
        thread = self.new(target, *args, threadTimeout=self.timeout, **kwargs)
        self.threadDictionary[len(self.threadDictionary)] = thread


    def addResource(self, api, app):
        api.resource.manager.queue = self
        self.api = api


    def onHttpRequestCompletion(self, api, app):
        ...


    def onRun(self, api, app):
        ...


    def initialize(self, api, app):
        self.threadDictionaryHandler.run()
        self.threadDictionary.run()
        log.success(self.initialize, f'{ReflectionHelper.getClassName(self)} is running')


    def onShutdown(self, api, app):
        self.threadDictionary.kill()
        self.threadDictionaryHandler.kill()
        log.success(self.onShutdown, f'{ReflectionHelper.getClassName(self)} is successfuly closed')



from python_helper import Constant as c
from python_helper import ObjectHelper, StringHelper
from python_framework import Mapper, MapperMethod, ConverterStatic


from constant import SubscriptionConstant
import SubscriptionDto
import SubscriptionModel


@Mapper()
class SubscriptionModelMapper:

    @MapperMethod(requestClass=[[SubscriptionDto.SubscriptionRequestDto], str], responseClass=[[SubscriptionModel.SubscriptionModel]])
    def fromRequestDtoListToModelList(self, dtoList, modelList):
        self.overrideAllModelOriginKey(originKey, modelList)
        return modelList


    @MapperMethod(requestClass=[[SubscriptionModel.SubscriptionModel]], responseClass=[[SubscriptionDto.SubscriptionResponseDto]])
    def fromModelListToResponseDtoList(self, modelList, dtoList):
        return dtoList


    @MapperMethod(requestClass=[SubscriptionDto.SubscriptionRequestDto, str], responseClass=[SubscriptionModel.SubscriptionModel])
    def fromRequestDtoToModel(self, dto, originKey, model):
        self.overrideModelOriginKey(originKey, model)
        return model


    @MapperMethod(requestClass=[SubscriptionModel.SubscriptionModel], responseClass=[SubscriptionDto.SubscriptionResponseDto])
    def fromModelToResponseDto(self, model, dto):
        return dto


    @MapperMethod(requestClass=[SubscriptionModel.SubscriptionModel, SubscriptionDto.SubscriptionRequestDto])
    def overrideModel(self, model, dto):
        model.url = ConverterStatic.getValueOrDefault(dto.url, model.url)
        model.onErrorUrl = ConverterStatic.getValueOrDefault(dto.onErrorUrl, model.onErrorUrl)
        model.maxTries = ConverterStatic.getValueOrDefault(dto.maxTries, model.maxTries)
        model.backOff = ConverterStatic.getValueOrDefault(dto.backOff, model.backOff)


    @MapperMethod(requestClass=[str, SubscriptionModel.SubscriptionModel])
    def overrideModelOriginKey(self, originKey, model):
        model.originKey = ConverterStatic.getValueOrDefault(originKey, model.originKey)


    @MapperMethod(requestClass=[str, [SubscriptionModel.SubscriptionModel]])
    def overrideAllModelOriginKey(self, originKey, modelList):
        for model in modelList:
            self.overrideModelOriginKey(originKey, model)



from python_helper import Constant as c
from python_helper import ObjectHelper, StringHelper
from python_framework import Mapper, MapperMethod, EnumItem

import Emission
import EmissionDto
from enumeration.ModelState import ModelState
from enumeration.ModelStatus import ModelStatus


@Mapper()
class EmissionMapper:

    @MapperMethod(requestClass=[[EmissionDto.EmissionRequestDto]], responseClass=[[Emission.Emission]])
    def fromRequestDtoListToModelList(self, dtoList, modelList):
        self.overrideAllModelState(modelList, ModelState.INSTANTIATED)
        return modelList


    @MapperMethod(requestClass=[[Emission.Emission]], responseClass=[[EmissionDto.EmissionResponseDto]])
    def fromModelListToResponseDtoList(self, modelList, dtoList):
        return dtoList


    @MapperMethod(requestClass=[EmissionDto.EmissionRequestDto], responseClass=[Emission.Emission])
    def fromRequestDtoToModel(self, dto, model):
        self.overrideModelState(model, ModelState.INSTANTIATED)
        return model


    @MapperMethod(requestClass=[Emission.Emission], responseClass=[EmissionDto.EmissionResponseDto])
    def fromModelToResponseDto(self, model, dto):
        return dto


    @MapperMethod(requestClass=[Emission.Emission, EnumItem])
    def overrideModelState(self, model, state):
        model.state = state


    @MapperMethod(requestClass=[[Emission.Emission], EnumItem])
    def overrideAllModelState(self, modelList, state):
        for model in modelList:
            self.overrideModelState(model, state)


    @MapperMethod(requestClass=[Emission.Emission, EnumItem])
    def overrideModelStatus(self, model, status):
        model.status = status
        self.overrideModelStateToModified(model)


    @MapperMethod(requestClass=[[Emission.Emission], EnumItem])
    def overrideAllModelStatus(self, modelList, status):
        for model in modelList:
            self.overrideModelStatus(model, status)


    @MapperMethod(requestClass=[Emission.Emission])
    def overrideModelStateToModified(self, model):
        if ModelStatus.ACCEPTED == model.status:
            self.mapper.message.overrideModelState(model, ModelState.INSTANTIATED)
        else:
            self.mapper.message.overrideModelState(model, ModelState.MODIFIED)



from python_helper import Constant as c
from python_helper import ObjectHelper, StringHelper
from python_framework import Mapper, MapperMethod, EnumItem

import Message
import MessageDto
from enumeration.ModelState import ModelState
from enumeration.ModelStatus import ModelStatus


@Mapper()
class MessageMapper:

    @MapperMethod(requestClass=[Message.Message], responseClass=[MessageDto.MessageCreationResponseDto])
    def fromModelToCreationResponseDto(self, model, dto):
        return dto


    @MapperMethod(requestClass=[[MessageDto.MessageRequestDto], str], responseClass=[[Message.Message]])
    def fromRequestDtoListToModelList(self, dtoList, originKey, modelList):
        for model in modelList:
            self.handleInitializationSpecialCases(originKey, model)
        return modelList


    @MapperMethod(requestClass=[[Message.Message]], responseClass=[[MessageDto.MessageResponseDto]])
    def fromModelListToResponseDtoList(self, modelList, dtoList):
        return dtoList


    @MapperMethod(requestClass=[MessageDto.MessageRequestDto, str], responseClass=[Message.Message])
    def fromRequestDtoToModel(self, dto, originKey, model):
        self.handleInitializationSpecialCases(originKey, model)
        return model


    @MapperMethod(requestClass=[Message.Message], responseClass=[MessageDto.MessageResponseDto])
    def fromModelToResponseDto(self, model, dto):
        return dto


    @MapperMethod(requestClass=[Message.Message, EnumItem])
    def overrideModelState(self, model, state):
        model.state = state


    @MapperMethod(requestClass=[[Message.Message], EnumItem])
    def overrideAllModelState(self, modelList, state):
        for model in modelList:
            self.overrideModelState(model, state)


    @MapperMethod(requestClass=[Message.Message, EnumItem])
    def overrideModelStatus(self, model, status):
        model.status = status
        self.overrideModelStateToModified(model)


    @MapperMethod(requestClass=[[Message.Message], EnumItem])
    def overrideAllModelStatus(self, modelList, status):
        for model in modelList:
            self.overrideModelStatus(model, status)


    @MapperMethod(requestClass=[Message.Message])
    def overrideModelStateToModified(self, model):
        if ModelStatus.ACCEPTED == model.status:
            self.mapper.message.overrideModelState(model, ModelState.INSTANTIATED)
        else:
            self.mapper.message.overrideModelState(model, ModelState.MODIFIED)


    @MapperMethod(requestClass=[str, Message.Message])
    def overrideModelOriginKey(self, originKey, model):
        model.originKey = originKey


    @MapperMethod(requestClass=[Message.Message])
    def overrideModelGroupKey(self, model):
        if ObjectHelper.isNone(model.groupKey):
            model.groupKey = model.key


    @MapperMethod(requestClass=[str, Message.Message])
    def handleInitializationSpecialCases(self, originKey, model):
        self.overrideModelStateToModified(model)
        self.overrideModelOriginKey(originKey, model)
        self.overrideModelGroupKey(model)



from python_helper import Constant as c
from python_helper import ObjectHelper
from python_framework import ConverterStatic, Serializer
from python_framework import SqlAlchemyProxy as sap

from ModelAssociation import EMISSION, MESSAGE
from constant import EmissionConstant, ModelConstant


class Emission:
    __memoryname__ = EMISSION.replace(Serializer.MODEL_SUFIX, c.BLANK)

    def __init__(self,
        queueKey = None,
        subscriptionKey = None,
        url = None,
        tries = None,
        onErrorUrl = None,
        onErrorTries = None,
        maxTries = None,
        backOff = None,
        status = None,
        state = None,
        message = None,
        history = None
    ):
        self.queueKey = queueKey
        self.subscriptionKey = subscriptionKey
        self.url = url
        self.tries = ConverterStatic.getValueOrDefault(tries, EmissionConstant.ZERO_TRIES)
        self.onErrorUrl = onErrorUrl
        self.onErrorTries = ConverterStatic.getValueOrDefault(onErrorTries, EmissionConstant.ZERO_TRIES)
        self.maxTries = ConverterStatic.getValueOrDefault(maxTries, EmissionConstant.DEFAULT_MAX_TRIES)
        self.backOff = ConverterStatic.getValueOrDefault(backOff, EmissionConstant.DEFAULT_BACKOFF)
        self.status = ConverterStatic.getValueOrDefault(status, ModelConstant.DEFAULT_STATUS)
        self.state = ConverterStatic.getValueOrDefault(state, ModelConstant.DEFAULT_STATE)
        self.setMessage(message)
        self.setHistory(history)


    def setMessage(self, message):
        self.message = message


    def getMessage(self, message=None):
        if ObjectHelper.isNotNone(message):
            self.setMessage(message)
        return self.message


    def getMessageKey(self, message=None):
        self.getMessage(message=message)
        return None if ObjectHelper.isNone(self.message) else self.message.key


    def getGroupKey(self, message=None):
        self.getMessage(message=message)
        return None if ObjectHelper.isNone(self.message) else self.message.groupKey


    def getOriginKey(self, message=None):
        self.getMessage(message=message)
        return None if ObjectHelper.isNone(self.message) else self.message.originKey


    def getHeaders(self):
        return dict() if ObjectHelper.isNone(self.message) else ConverterStatic.getValueOrDefault(self.message.headers, dict())


    def getContent(self):
        return None if ObjectHelper.isNone(self.message) else self.message.content


    def setHistory(self, history):
        if ObjectHelper.isNone(history):
            self.history = []
        elif ObjectHelper.isNotList(history):
            self.history = [str(history)]
        else:
            self.history = [
                str(h) for h in ConverterStatic.getValueOrDefault(history, [])
            ]
        # self.history = []


    def addHistory(self, history):
        self.history.append(str(history))
        # self.history = []


    def __repr__(self):
        return f'{self.__memoryname__}(queueKey={self.queueKey}, subscriptionKey={self.subscriptionKey}, messageKey={self.getMessageKey()}, groupKey={self.getGroupKey()}, originKey={self.getOriginKey()}, tries={self.tries}, onErrorTries={self.onErrorTries}, maxTries={self.maxTries}, backOff={self.backOff}, status={self.status}, state={self.state})'



from python_helper import log, ObjectHelper
from python_framework import Service, ServiceMethod

import MessageDto
import MessageModel, Message, EmissionModel
from enumeration.ModelStatus import ModelStatus
from enumeration.ModelState import ModelState


LOG_LEVEL = log.STATUS


@Service()
class MessageModelService:

    @ServiceMethod(requestClass=[[Message.Message]])
    def createOrUpdateAll(self, messageList):
        if ObjectHelper.isEmpty(messageList):
            return []
        self.mapper.message.overrideAllModelState(messageList, ModelState.PERSISTED)
        modelList = self.mapper.messageModel.fromRequestDtoListToModelList(messageList)
        log.prettyPython(self.createOrUpdateAll, f'Creating or updating queued messages', modelList, logLevel=LOG_LEVEL)
        existingModelList = self.findAllModelByKeyIn(list({model.key for model in modelList}))
        existingModelDictionary = {
            model.key: model for model in existingModelList
        }
        for model in modelList:
            if model.key in existingModelDictionary:
                self.mapper.messageModel.overrideModel(
                    existingModelDictionary.get(model.key),
                    model
                )
            else:
                existingModelList.append(model)
        return self.persistAll(existingModelList)


    @ServiceMethod(requestClass=[Message.Message])
    def findOrCreateModel(self, model):
        existingModel = self.findByKey(model.key)
        if ObjectHelper.isNotNone(existingModel):
            return existingModel
        return self.persistAll([existingModel])[0]


    @ServiceMethod(requestClass=[str])
    def existsByKey(self, key):
        return self.repository.messageModel.existsByKey(key)


    @ServiceMethod(requestClass=[[str]])
    def findAllModelByKeyIn(self, modelKeyList):
        return self.repository.messageModel.findAllByKeyIn(modelKeyList)


    @ServiceMethod(requestClass=[[MessageModel.MessageModel]])
    def persistAll(self, modelList):
        self.mapper.messageModel.overrideAllModelState(modelList, ModelState.PERSISTED)
        log.prettyPython(self.createOrUpdateAll, f'Persisting queued messages', modelList, logLevel=LOG_LEVEL)
        return self.repository.messageModel.saveAll(modelList)



from python_framework import Serializer, HttpStatus, JwtConstant
from MessageEmitterAnnotation import MessageEmitter
from MessageEmitterAnnotation import MessageEmitterMethod

from config import MessageConfig
import MessageDto
import Message


@MessageEmitter(
    url = MessageConfig.EMITTER_URL,
    timeout = MessageConfig.EMITTER_TIMEOUT,
    headers = {
        'Content-Type': 'application/json',
        JwtConstant.DEFAULT_JWT_API_KEY_HEADER_NAME: f'Bearer {MessageConfig.MESSAGE_API_KEY}'
    }
    , muteLogs = False
    , logRequest = True
    , logResponse = True
)
class MessageEmitter:

    @MessageEmitterMethod(
        queueKey = MessageConfig.SOME_UNIQUE_QUEUE_KEY,
        requestClass=[MessageDto.MessageRequestDto],
        responseClass=[MessageDto.MessageRequestDto]
        , logRequest = True
        , logResponse = True
    )
    def send(self, dto):
        return self.emit(body=dto)



from python_helper import Constant as c
from python_helper import ObjectHelper, StringHelper
from python_framework import Mapper, MapperMethod, EnumItem, ConverterStatic

import MessageModel
import Message
from enumeration.ModelState import ModelState
from enumeration.ModelStatus import ModelStatus


@Mapper()
class MessageModelMapper:

    @MapperMethod(requestClass=[[Message.Message]], responseClass=[[MessageModel.MessageModel]])
    def fromRequestDtoListToModelList(self, dtoList, modelList):
        for model in modelList:
            self.handleInitializationSpecialCases(model)
        return modelList


    @MapperMethod(requestClass=[[MessageModel.MessageModel]], responseClass=[[Message.Message]])
    def fromModelListToResponseDtoList(self, modelList, dtoList):
        return dtoList


    @MapperMethod(requestClass=[Message.Message], responseClass=[MessageModel.MessageModel])
    def fromRequestDtoToModel(self, dto, model):
        self.handleInitializationSpecialCases(model)
        return model


    @MapperMethod(requestClass=[MessageModel.MessageModel], responseClass=[Message.Message])
    def fromModelToResponseDto(self, model, dto):
        return dto


    @MapperMethod(requestClass=[MessageModel.MessageModel, MessageModel.MessageModel])
    def overrideModel(self, modelToOverride, model):
        self.handleInitializationSpecialCases(modelToOverride)
        modelToOverride.status = model.status


    @MapperMethod(requestClass=[MessageModel.MessageModel, EnumItem])
    def overrideModelState(self, model, state):
        model.state = state
        ConverterStatic.overrideDateData(model)


    @MapperMethod(requestClass=[[MessageModel.MessageModel], EnumItem])
    def overrideAllModelState(self, modelList, state):
        for model in modelList:
            self.overrideModelState(model, state)


    @MapperMethod(requestClass=[MessageModel.MessageModel, EnumItem])
    def overrideModelStatus(self, model, status):
        self.handleInitializationSpecialCases(model)
        model.status = status


    @MapperMethod(requestClass=[[MessageModel.MessageModel], EnumItem])
    def overrideAllModelStatus(self, modelList, status):
        for model in modelList:
            self.overrideModelStatus(model, status)


    @MapperMethod(requestClass=[MessageModel.MessageModel])
    def overrideModelStateToModified(self, model):
        if ModelStatus.ACCEPTED == model.status:
            self.mapper.message.overrideModelState(model, ModelState.INSTANTIATED)
        else:
            self.mapper.message.overrideModelState(model, ModelState.MODIFIED)


    @MapperMethod(requestClass=[MessageModel.MessageModel])
    def overrideModelGroupKey(self, model):
        if ObjectHelper.isNone(model.groupKey):
            model.groupKey = model.key


    @MapperMethod(requestClass=[MessageModel.MessageModel])
    def handleInitializationSpecialCases(self, model):
        self.overrideModelStateToModified(model)
        self.overrideModelGroupKey(model)



from python_helper import Constant as c
from python_helper import ObjectHelper, StringHelper
from python_framework import ConverterStatic, Serializer
from python_framework import SqlAlchemyProxy as sap

from ModelAssociation import MESSAGE, EMISSION, MODEL
from constant import ModelConstant
from util import ModelUtil


class MessageModel(MODEL):
    __tablename__ = MESSAGE

    id = sap.Column(sap.Integer(), sap.Sequence(f'{__tablename__}{sap.ID}{sap.SEQ}'), primary_key=True)
    key = sap.Column(sap.String(sap.MEDIUM_STRING_SIZE), nullable=False, unique=True)
    queueKey = sap.Column(sap.String(sap.MEDIUM_STRING_SIZE), nullable=False)
    groupKey = sap.Column(sap.String(sap.MEDIUM_STRING_SIZE), nullable=False)
    originKey = sap.Column(sap.String(sap.MEDIUM_STRING_SIZE), nullable=False)
    content = sap.Column(sap.String(65_536))
    status = sap.Column(sap.String(sap.LITTLE_STRING_SIZE), nullable=False, default=ModelConstant.DEFAULT_STATUS)
    state = sap.Column(sap.String(sap.LITTLE_STRING_SIZE), nullable=False, default=ModelConstant.DEFAULT_STATE)

    history = sap.Column(sap.String(65_536))

    createdAt = sap.Column(sap.DateTime, nullable=False)
    updatedAt = sap.Column(sap.DateTime, nullable=False)


    def __init__(self,
        id = None,
        key = None,
        queueKey = None,
        groupKey = None,
        originKey = None,
        content = None,
        status = None,
        state = None,
        emissionList = None,
        history = None,
        createdAt = None,
        updatedAt = None
    ):
        self.id = id
        self.key = key
        self.queueKey = queueKey
        self.groupKey = groupKey
        self.originKey = originKey
        self.content = Serializer.jsonifyIt(content)
        self.status = ConverterStatic.getValueOrDefault(status, ModelConstant.DEFAULT_STATUS)
        self.state = ConverterStatic.getValueOrDefault(state, ModelConstant.DEFAULT_STATE)
        self.setHistory(history) ###- ModelUtil.getOneToManyData(history)

        self.createdAt = createdAt
        self.updatedAt = updatedAt
        ConverterStatic.overrideDateData(self)


    def setHistory(self, history):
        self.history = str(history)


    def addHistory(self, history):
        self.history = str(history)


    def __repr__(self):
        return f'{self.__tablename__}(id={self.id}, key={self.key}, queueKey={self.queueKey}, status={self.status}, state={self.state})'



from python_framework import Repository

import MessageModel


@Repository(model = MessageModel.MessageModel)
class MessageModelRepository:

    def save(self, model) :
        return self.repository.saveAndCommit(model)

    def saveAll(self, modelList):
        return self.repository.saveAllAndCommit(modelList)

    def findAll(self) :
        return self.repository.findAllAndCommit(self.model)

    def findByKey(self, key) :
        if self.existsByKey(key) :
            return self.repository.findByKeyAndCommit(key, self.model)

    def findAllByKeyIn(self, keyList) :
        modelList = self.repository.session.query(self.model).filter(self.model.key.in_(keyList)).all()
        self.repository.session.commit()
        return modelList

    def existsByKey(self, key) :
        return self.repository.existsByKeyAndCommit(key, self.model)

    def notExistsByKey(self, key) :
        return not self.existsByKey(key)

    def deleteByKey(self, key):
        self.repository.deleteByKeyAndCommit(key, self.model)

    def findById(self, id) :
        if self.existsById(id) :
            return self.repository.findByIdAndCommit(id, self.model)

    def findAllByIdIn(self, idList) :
        modelList = self.repository.session.query(self.model).filter(self.model.id.in_(idList)).all()
        self.repository.session.commit()
        return modelList

    def existsById(self, id) :
        return self.repository.existsByIdAndCommit(id, self.model)

    def notExistsById(self, id) :
        return not self.existsById(id)

    def deleteById(self, id):
        self.repository.deleteByIdAndCommit(id, self.model)



from python_helper import ObjectHelper
from python_framework import Service, ServiceMethod, GlobalException

from util import AuditoryUtil
import SubscriptionDto


@Service()
class SubscriptionModelService:

    @ServiceMethod()
    def findAllByOrigin(self):
        return self.mapper.subscriptionModel.fromModelListToResponseDtoList(
            self.repository.subscriptionModel.findAllByOriginKey(AuditoryUtil.getApiKeyIdentity(service=self))
        )


    @ServiceMethod(requestClass=[SubscriptionDto.SubscriptionRequestDto])
    def createOrUpdate(self, dto):
        raise GlobalException(message='Method not implemented', status=HttpStatus.I_M_A_TEAPOT)


    @ServiceMethod(requestClass=[str])
    def findAllModelByQueueKey(self, queueKey):
        modelList = self.repository.subscriptionModel.findAllByQueueKey(queueKey)
        self.validator.subscriptionModel.validateIsAtLeastOne(modelList, queueKey)
        return modelList


    @ServiceMethod(requestClass=[[str]])
    def findAllModelByQueueKeyIn(self, queueKeyList):
        raise Exception('You should never make such a question')


    @ServiceMethod(requestClass=[str])
    def existsByKey(self, key):
        return self.repository.message.existsByKey(key)


    @ServiceMethod(requestClass=[str])
    def findModelByKey(self, key):
        modelList = self.findAllModelByKey(key)
        self.validator.queueModel.validateIsExactlyOne(modelList)
        return modelList[0]


    @ServiceMethod(requestClass=[str])
    def findAllModelByKey(self, key):
        return self.repository.subscriptionModel.findAllByKey(key)



from python_framework import Controller, ControllerMethod, HttpStatus

from enumeration.AccessDomain import AccessDomain
import SubscriptionDto


@Controller(url = '/subscription', tag='Subscription', description='Subscription controller')
class SubscriptionModelController:

    @ControllerMethod(url = '/',
        apiKeyRequired = [AccessDomain.API],
        responseClass = [[SubscriptionDto.SubscriptionResponseDto]]
        # , logRequest = True
        # , logResponse = True
    )
    def get(self):
        return self.service.subscriptionModel.findAllByOrigin(), HttpStatus.OK


    @ControllerMethod(url = '/',
        apiKeyRequired = [AccessDomain.API],
        requestClass = [SubscriptionDto.SubscriptionRequestDto],
        responseClass = [SubscriptionDto.SubscriptionResponseDto]
        # , logRequest = True
        # , logResponse = True
    )
    def post(self, dto):
        return self.service.subscriptionModel.createOrUpdate(dto), HttpStatus.CREATED



from python_helper import log
from globals import getGlobalsInstance
globalsInstance = getGlobalsInstance()


MESSAGE_API_KEY = globalsInstance.getSetting('queue.message.api-key')

LISTENER_TIMEOUT = globalsInstance.getSetting('queue.message.listener.timeout')

EMITTER_URL = globalsInstance.getSetting('queue.message.emitter.url')
EMITTER_TIMEOUT = globalsInstance.getSetting('queue.message.emitter.timeout')
SOME_UNIQUE_QUEUE_KEY = globalsInstance.getSetting('queue.message.emitter.queue-key.test-1')



from python_helper import Constant as c
from python_helper import ObjectHelper, StringHelper
from python_framework import Mapper, MapperMethod, EnumItem, ConverterStatic

import EmissionModel
import Emission
from enumeration.ModelState import ModelState
from enumeration.ModelStatus import ModelStatus


@Mapper()
class EmissionModelMapper:

    @MapperMethod(requestClass=[[Emission.Emission]], responseClass=[[EmissionModel.EmissionModel]])
    def fromRequestDtoListToModelList(self, dtoList, modelList):
        for model, dto in zip(modelList, dtoList):
            self.overrideMessageKey(model, dto)
        return modelList


    @MapperMethod(requestClass=[[EmissionModel.EmissionModel]], responseClass=[[Emission.Emission]])
    def fromModelListToResponseDtoList(self, modelList, dtoList):
        raise Exception('This shoud not be called due history raw implmentation')
        return dtoList


    @MapperMethod(requestClass=[Emission.Emission], responseClass=[EmissionModel.EmissionModel])
    def fromRequestDtoToModel(self, dto, model):
        self.overrideMessageKey(model, dto)
        return model


    @MapperMethod(requestClass=[EmissionModel.EmissionModel], responseClass=[Emission.Emission])
    def fromModelToResponseDto(self, model, dto):
        raise Exception('This shoud not be called due history raw implmentation')
        return dto


    @MapperMethod(requestClass=[EmissionModel.EmissionModel, Emission.Emission])
    def overrideMessageKey(self, toOverridemodel, emission):
        toOverridemodel.messageKey = emission.getMessageKey()
        toOverridemodel.groupKey = emission.getGroupKey()
        toOverridemodel.originKey = emission.getOriginKey()
        toOverridemodel.updateKey()
        self.overrideModelStateToModified(toOverridemodel)


    @MapperMethod(requestClass=[EmissionModel.EmissionModel, EmissionModel.EmissionModel])
    def overrideModel(self, toOverridemodel, model):
        toOverridemodel.url = model.url
        toOverridemodel.tries = model.tries
        toOverridemodel.onErrorUrl = model.onErrorUrl
        toOverridemodel.onErrorTries = model.onErrorTries
        toOverridemodel.maxTries = model.maxTries
        toOverridemodel.backOff = model.backOff
        toOverridemodel.status = model.status
        toOverridemodel.setHistory(model.history)
        self.overrideModelStateToModified(toOverridemodel)


    @MapperMethod(requestClass=[EmissionModel.EmissionModel, EnumItem])
    def overrideModelState(self, model, state):
        model.state = state


    @MapperMethod(requestClass=[[EmissionModel.EmissionModel], EnumItem])
    def overrideAllModelState(self, modelList, state):
        for model in modelList:
            self.overrideModelState(model, state)
            ConverterStatic.overrideDateData(model)


    @MapperMethod(requestClass=[EmissionModel.EmissionModel, EnumItem])
    def overrideModelStatus(self, model, status):
        model.status = status
        self.overrideModelStateToModified(model)


    @MapperMethod(requestClass=[[EmissionModel.EmissionModel], EnumItem])
    def overrideAllModelStatus(self, modelList, status):
        for model in modelList:
            self.overrideModelStatus(model, status)


    @MapperMethod(requestClass=[EmissionModel.EmissionModel])
    def overrideModelStateToModified(self, model):
        if ModelStatus.ACCEPTED == model.status:
            self.mapper.message.overrideModelState(model, ModelState.INSTANTIATED)
        else:
            self.mapper.message.overrideModelState(model, ModelState.MODIFIED)



from python_helper import log, ObjectHelper
from python_framework import Service, ServiceMethod

import MessageDto
import Message
from enumeration.ModelStatus import ModelStatus
from enumeration.ModelState import ModelState
from util import AuditoryUtil


LOG_LEVEL = log.STATUS


@Service()
class MessageService:

    @ServiceMethod(requestClass=[MessageDto.MessageRequestDto])
    def accept(self, dto):
        self.validator.messageModel.validateDoesNotExists(dto)
        model = self.mapper.message.fromRequestDtoToModel(dto, AuditoryUtil.getApiKeyIdentity(service=self))
        self.mapper.message.overrideModelStatus(model, ModelStatus.ACCEPTED)
        log.prettyPython(self.accept, f'Accepting new queued message', model, logLevel=LOG_LEVEL)
        return self.mapper.message.fromModelToCreationResponseDto(self.service.memory.acceptMessage(model))


    @ServiceMethod()
    def sendAllAcceptedFromOneQueue(self):
        modelList = self.service.memory.getAllAcceptedMessagesFromOneQueue()
        if ObjectHelper.isEmpty(modelList):
            return []
        self.mapper.message.overrideAllModelStatus(modelList, ModelStatus.PROCESSING)
        log.prettyPython(self.sendAllAcceptedFromOneQueue, f'Processing queued messages', modelList, logLevel=LOG_LEVEL)
        try:
            self.validator.message.validateAllBelongsToTheSameQueue(modelList)
            queue = self.service.queueModel.findModelByKey(modelList[0].queueKey)
            self.service.emission.acceptWithoutValidation(modelList, queue)
            self.mapper.message.overrideAllModelStatus(modelList, ModelStatus.PROCESSED)
        except Exception as exception:
            self.mapper.message.overrideAllModelStatus(modelList, ModelStatus.PROCESSED_WITH_ERRORS)
            for model in modelList:
                model.addHistory(exception)
            raise exception


    @ServiceMethod()
    def updateAllModifiedFromMemory(self):
        modelList = self.service.memory.getAllModifiedMessages()
        return self.service.messageModel.createOrUpdateAll(modelList)



from python_framework import SchedulerType
from python_framework import Scheduler, SchedulerMethod, WeekDay, WeekDayConstant


@Scheduler(muteLogs=True)
class MemoryScheduler:

    @SchedulerMethod(SchedulerType.INTERVAL, seconds=15, instancesUpTo=2)
    def removeAllProcessedMessagesAndEmissions(self) :
        self.service.memory.removeAllProcessedMessagesAndEmissions()



from python_helper import Constant as c
from python_helper import ReflectionHelper, ObjectHelper, log, Function, StringHelper
from python_framework import (
    FlaskManager,
    ConverterStatic,
    Listener,
    ListenerMethod,
    FlaskUtil,
    Serializer
)


DEFAUTL_RESOURCE_ENABLED = False
DEFAUTL_RESOURCE_MUTE_LOGS = False
DEFAUTL_RESOURCE_METHOD_ENABLED = True
DEFAUTL_RESOURCE_METHOD_MUTE_LOGS = False
DEFAULT_TIMEOUT = 30


def initializeComunicationLayerResource(
    resourceInstance = None,
    api = None,
    enabled = None,
    muteLogs = None,
    logRequest = False,
    logResponse = False,
    timeout = None,
    resourceEnabledConfigKey = None,
    resourceMuteLogsConfigKey = None,
    resourceTimeoutConfigKey = None,
    defaultEnabled = DEFAUTL_RESOURCE_ENABLED,
    defaultMuteLogs = DEFAUTL_RESOURCE_MUTE_LOGS,
    defaultTimeout = DEFAULT_TIMEOUT
):
    api = FlaskUtil.retrieveApiInstance(apiInstance=api, arguments=(resourceInstance,))
    resourceInstance.enabled = enabled and ConverterStatic.getValueOrDefault(
        api.globals.getApiSetting(resourceEnabledConfigKey),
        defaultEnabled
    )
    resourceInstance.muteLogs = muteLogs or ConverterStatic.getValueOrDefault(
        api.globals.getApiSetting(resourceMuteLogsConfigKey),
        defaultMuteLogs
    )
    resourceInstance.timeout = ConverterStatic.getValueOrDefault(
        timeout,
        ConverterStatic.getValueOrDefault(
            api.globals.getApiSetting(resourceTimeoutConfigKey),
            defaultTimeout
        )
    )
    resourceInstance.logRequest = logRequest
    resourceInstance.logResponse = logResponse
    resourceInstance.globals = api.globals
    resourceInstance.service = api.resource.service
    resourceInstance.defaultEnabled = defaultEnabled
    resourceInstance.defaultMuteLogs = defaultMuteLogs
    resourceInstance.defaultTimeout = defaultTimeout


class InnerMethodWrapperManager:

    def __init__(
        self,
        resourceInstanceMethodArguments = None,
        wrapperType = None,
        resourceInstanceMethod = None,
        timeout = None,
        enabled = None,
        muteLogs = None,
        resourceEnabled = None,
        resourceMuteLogs = None,
        logRequest = False,
        logResponse = False,
        resourceTypeName = None,
        resourceEnabledConfigKey = None,
        resourceMuteLogsConfigKey = None,
        resourceTimeoutConfigKey = None,
        defaultEnabled = DEFAUTL_RESOURCE_METHOD_ENABLED,
        defaultMuteLogs = DEFAUTL_RESOURCE_METHOD_MUTE_LOGS,
        defaultTimeout = DEFAULT_TIMEOUT,
        **methodKwargs
    ):
        log.wrapper(wrapperType, f'''wrapping {resourceInstanceMethod.__name__}''')
        self.api = FlaskManager.getApi()

        self.resourceInstance = None
        self.resourceInstanceMethod = resourceInstanceMethod
        self.resourceTypeName = resourceTypeName
        self.methodClassName = ReflectionHelper.getMethodClassName(self.resourceInstanceMethod)
        self.methodName = ReflectionHelper.getName(self.resourceInstanceMethod)
        resourceInstanceName = self.methodClassName[:-len(self.resourceTypeName)]
        self.resourceInstanceName = f'{resourceInstanceName[0].lower()}{resourceInstanceName[1:]}'
        self.id = methodKwargs.get('id', f'{self.methodClassName}{c.DOT}{self.methodName}')
        self.defaultEnabled = defaultEnabled
        self.defaultMuteLogs = defaultMuteLogs
        self.defaultTimeout = defaultTimeout
        self.enabled = enabled and ConverterStatic.getValueOrDefault(
            self.api.globals.getApiSetting(resourceEnabledConfigKey),
            self.defaultEnabled
        )
        self.muteLogs = muteLogs or ConverterStatic.getValueOrDefault(
            self.api.globals.getApiSetting(resourceMuteLogsConfigKey),
            self.defaultMuteLogs
        )
        self.timeout = ConverterStatic.getValueOrDefault(
            timeout,
            ConverterStatic.getValueOrDefault(
                self.api.globals.getApiSetting(resourceTimeoutConfigKey),
                self.defaultTimeout
            )
        )
        self.logRequest = logRequest
        self.logResponse = logResponse
        self.resourceInstance = self.updateResourceInstance(ConverterStatic.getValueOrDefault(resourceInstanceMethodArguments, list()))


    def shouldLogRequest(self):
        return self.resourceInstance.logRequest and self.logRequest


    def shouldLogResponse(self):
        return self.resourceInstance.logResponse and self.logResponse


    def updateResourceInstance(self, args):
        if ObjectHelper.isNone(self.resourceInstance):
            if ObjectHelper.isEmpty(args):
                try:
                    self.resourceInstance = FlaskManager.getResourceSelf(
                        self.api,
                        self.resourceTypeName,
                        self.resourceInstanceName
                    )
                except Exception as exception:
                    log.log(self.updateResourceInstance, f'Not possible to get "{self.resourceInstanceName}" resource instance. Make sure to add it in method usage scope', exception=exception, muteStackTrace=True)
            else :
                self.resourceInstance = args[0]
        try :
            resourceInstanceEnabled = ConverterStatic.getValueOrDefault(self.resourceInstance.enabled, self.defaultEnabled)
            resourceInstanceMuteLogs = ConverterStatic.getValueOrDefault(self.resourceInstance.muteLogs, self.defaultMuteLogs)
            self.timeout = ConverterStatic.getValueOrDefault(self.timeout, ConverterStatic.getValueOrDefault(self.resourceInstance.timeout, self.defaultTimeout))
            self.enabled = resourceInstanceEnabled and ConverterStatic.getValueOrDefault(self.enabled, self.defaultEnabled)
            self.muteLogs = resourceInstanceMuteLogs or ConverterStatic.getValueOrDefault(self.muteLogs, self.defaultMuteLogs)
            self.logRequest = self.logRequest and ConverterStatic.getValueOrDefault(self.resourceInstance.logRequest, False)
            self.logResponse = self.logResponse and ConverterStatic.getValueOrDefault(self.resourceInstance.logResponse, False)
        except Exception as exception:
            log.log(self.updateResourceInstance, f'Not possible to update "{self.resourceInstanceName}" resource instance configurations properly. Make sure to do it within method usage scope', exception=exception, muteStackTrace=True)
        return self.resourceInstance


    def addResourceInFrontOfArgs(self, args):
        return FlaskManager.getArgumentInFrontOfArgs(args, self.updateResourceInstance(tuple()))



import json
from flask import copy_current_request_context

from python_helper import Constant as c
from python_helper import ReflectionHelper, ObjectHelper, log, Function, StringHelper
from python_framework import (
    FlaskManager,
    GlobalException,
    ConverterStatic,
    FlaskUtil,
    Serializer,
    ConfigurationKeyConstant,
    EncapsulateItWithGlobalException,
    OpenApiManager,
    HttpStatus,
    HttpDomain,
    LogConstant
)

try:
    import MessageConstant
    import AnnotationUtil
    import MessageDto
except:
    from queue_manager_api.api.src.constant import MessageConstant
    from queue_manager_api.api.src.util import AnnotationUtil
    from queue_manager_api.api.src.dto import MessageDto


DEFAULT_TIMEOUT = 2


@Function
def MessageListener(
    *resourceArgs,
    url = c.SLASH,
    timeout = DEFAULT_TIMEOUT,
    responseHeaders = None,
    logRequest = False,
    logResponse = False,
    enabled = None,
    muteLogs = None,
    **resourceKwargs
):
    def Wrapper(OuterClass, *outterArgs, **outterKwargs):
        resourceUrl = url
        resourceTimeout = timeout
        resourceLogRequest = logRequest
        resourceLogResponse = logResponse
        resourceEnabled = enabled
        resourceMuteLogs = muteLogs
        resourceInstanceResponseHeaders = responseHeaders
        log.wrapper(MessageListener, f'''wrapping {OuterClass.__name__}(*{outterArgs}, **{outterKwargs})''')
        api = FlaskManager.getApi()
        class InnerClass(OuterClass):
            url = resourceUrl
            responseHeaders = resourceInstanceResponseHeaders
            logRequest = resourceLogRequest
            logResponse = resourceLogResponse
            enabled = resourceEnabled
            muteLogs = resourceMuteLogs
            def __init__(self, *args, **kwargs):
                log.wrapper(OuterClass, f'in {InnerClass.__name__}.__init__(*{args},**{kwargs})')
                OuterClass.__init__(self, *args,**kwargs)
                AnnotationUtil.initializeComunicationLayerResource(
                    resourceInstance = self,
                    api = api,
                    timeout = resourceTimeout,
                    enabled = resourceEnabled,
                    muteLogs = resourceMuteLogs,
                    logRequest = resourceLogRequest,
                    logResponse = resourceLogResponse,
                    resourceEnabledConfigKey = ConfigurationKeyConstant.API_LISTENER_ENABLE,
                    resourceMuteLogsConfigKey = ConfigurationKeyConstant.API_LISTENER_MUTE_LOGS,
                    resourceTimeoutConfigKey = ConfigurationKeyConstant.API_LISTENER_TIMEOUT
                )
                self.manager = api.resource.manager.queue
        ReflectionHelper.overrideSignatures(InnerClass, OuterClass)
        return InnerClass
    return Wrapper


@Function
def MessageListenerMethod(
    *methodArgs,
    url = c.SLASH,
    timeout = DEFAULT_TIMEOUT,
    requestHeaderClass = None,
    requestParamClass = None,
    requestClass = None,
    responseClass = None,
    responseHeaders = None,
    roleRequired = None,
    apiKeyRequired = None,
    contextRequired = None,
    consumes = OpenApiManager.DEFAULT_CONTENT_TYPE,
    produces = OpenApiManager.DEFAULT_CONTENT_TYPE,
    logRequest = True,
    logResponse = True,
    enabled = True,
    muteLogs = False,
    muteStacktraceOnBusinessRuleException = True,
    runInAThread = False,
    **methodKwargs
):
    def innerMethodWrapper(resourceInstanceMethod, *innerMethodArgs, **innerMethodKwargs):
        resourceInstanceMethodUrl = url
        wrapperManager = AnnotationUtil.InnerMethodWrapperManager(
            wrapperType = MessageListenerMethod,
            resourceInstanceMethod = resourceInstanceMethod,
            timeout = timeout,
            enabled = enabled,
            muteLogs = muteLogs,
            logRequest = logRequest,
            logResponse = logResponse,
            resourceTypeName = FlaskManager.KW_LISTENER_RESOURCE,
            resourceEnabledConfigKey = ConfigurationKeyConstant.API_LISTENER_ENABLE,
            resourceMuteLogsConfigKey = ConfigurationKeyConstant.API_LISTENER_MUTE_LOGS,
            resourceTimeoutConfigKey = ConfigurationKeyConstant.API_LISTENER_TIMEOUT
        )
        resourceInstanceMethodMuteStacktraceOnBusinessRuleException = muteStacktraceOnBusinessRuleException
        resourceInstanceMethodRunInAThread = runInAThread
        listenerUrl = f'{wrapperManager.api.baseUrl}{resourceInstanceMethodUrl}'
        if listenerUrl.endswith(c.SLASH):
            listenerUrl = listenerUrl[:-1]
        @wrapperManager.api.app.route(listenerUrl, methods=[HttpDomain.Verb.POST], endpoint=f'{resourceInstanceMethod.__qualname__}')
        def innerResourceInstanceMethod(*args, **kwargs):
            args = wrapperManager.addResourceInFrontOfArgs(args)
            messageAsJson = FlaskUtil.safellyGetRequestBody()
            completeResponse = None
            if not wrapperManager.muteLogs:
                requestUrl = FlaskUtil.safellyGetUrl()
                requestVerb = FlaskUtil.safellyGetVerb()
                log.info(wrapperManager.resourceInstanceMethod, f'''{LogConstant.LISTENER_SPACE}{requestVerb}{c.SPACE_DASH_SPACE}{requestUrl}''')
                if wrapperManager.shouldLogRequest():
                    try:
                        messageCreationRequestDto = Serializer.getObjectAsDictionary(Serializer.convertFromJsonToObject(messageAsJson, MessageDto.MessageCreationRequestDto))
                        log.prettyPython(wrapperManager.resourceInstanceMethod, f'{LogConstant.LISTENER_SPACE}Message data', messageCreationRequestDto, logLevel=log.INFO)
                    except Exception as exception:
                        log.failure(innerResourceInstanceMethod, 'Not possible to log message data properly', exception)
            if not wrapperManager.enabled:
                completeResponse = FlaskManager.getCompleteResponseByException(
                    GlobalException(logMessage='This resource is temporarily disabled', status=HttpStatus.SERVICE_UNAVAILABLE),
                    wrapperManager.resourceInstance,
                    wrapperManager.resourceInstanceMethod,
                    resourceInstanceMethodMuteStacktraceOnBusinessRuleException,
                    context = HttpDomain.LISTENER_CONTEXT
                )
            elif ObjectHelper.isEmpty(messageAsJson):
                completeResponse = FlaskManager.getCompleteResponseByException(
                    GlobalException(message='Content cannot be empty', logResponse=f'Content: {messageContent}', status=HttpStatus.BAD_REQUEST),
                    wrapperManager.resourceInstance,
                    wrapperManager.resourceInstanceMethod,
                    resourceInstanceMethodMuteStacktraceOnBusinessRuleException,
                    context = HttpDomain.LISTENER_CONTEXT
                )
            else:
                listennerArgs = (
                    args,
                    kwargs,
                    wrapperManager,
                    roleRequired,
                    apiKeyRequired,
                    contextRequired,
                    requestHeaderClass,
                    requestParamClass,
                    requestClass,
                    responseClass,
                    responseHeaders,
                    consumes,
                    resourceInstanceMethodMuteStacktraceOnBusinessRuleException,

                    requestUrl,
                    requestVerb,
                    FlaskUtil.safellyGetHeaders(),
                    FlaskUtil.safellyGetArgs(),
                    messageAsJson.get(MessageConstant.MESSAGE_CONTENT_KEY, {})
                )
                if resourceInstanceMethodRunInAThread:
                    wrapperManager.resourceInstance.manager.runInAThread(resolveListenerCallWithinAContext, *listennerArgs)
                else:
                    completeResponse = resolveListenerCall(*listennerArgs)
                if ObjectHelper.isEmpty(completeResponse) or HttpStatus.BAD_REQUEST < completeResponse[-1]:
                    completeResponse = [
                        MessageDto.MessageCreationResponseDto(
                            key = messageAsJson.get(MessageConstant.MESSAGE_KEY_KEY),
                            queueKey = messageAsJson.get(MessageConstant.MESSAGE_QUEUE_KEY_KEY),
                            groupKey = messageAsJson.get(MessageConstant.MESSAGE_GROUP_KEY)
                        ),
                        {},
                        HttpStatus.ACCEPTED
                    ]
            httpResponse = FlaskUtil.buildHttpResponse(completeResponse[1], completeResponse[0], completeResponse[-1].enumValue, produces)
            if wrapperManager.shouldLogResponse():
                try:
                    resourceMethodResponseStatus = completeResponse[-1]
                    log.prettyJson(
                        wrapperManager.resourceInstanceMethod,
                        LogConstant.LISTENER_RESPONSE,
                        {
                            'headers': FlaskUtil.safellyGetResponseHeaders(httpResponse),
                            'body': FlaskUtil.safellyGetFlaskResponseJson(httpResponse),
                            'status': resourceMethodResponseStatus.enumValue
                        },
                        condition = True,
                        logLevel = log.INFO
                    )
                except Exception as exception:
                    log.failure(innerResourceInstanceMethod, 'Not possible to log response properly', exception)

            return httpResponse
        ReflectionHelper.overrideSignatures(innerResourceInstanceMethod, wrapperManager.resourceInstanceMethod)
        innerResourceInstanceMethod.url = url
        innerResourceInstanceMethod.requestHeaderClass = requestHeaderClass
        innerResourceInstanceMethod.requestParamClass = requestParamClass
        innerResourceInstanceMethod.requestClass = requestClass
        innerResourceInstanceMethod.responseClass = responseClass
        innerResourceInstanceMethod.responseHeaders = responseHeaders
        innerResourceInstanceMethod.apiKeyRequired = apiKeyRequired
        innerResourceInstanceMethod.produces = produces
        innerResourceInstanceMethod.consumes = consumes
        innerResourceInstanceMethod.logRequest = logRequest
        innerResourceInstanceMethod.logResponse = logResponse
        innerResourceInstanceMethod.enabled = enabled
        innerResourceInstanceMethod.muteLogs = muteLogs
        innerResourceInstanceMethod.muteStacktraceOnBusinessRuleException = resourceInstanceMethodMuteStacktraceOnBusinessRuleException
        innerResourceInstanceMethod.runInAThread = resourceInstanceMethodRunInAThread
        return innerResourceInstanceMethod
    return innerMethodWrapper


###- https://flask.palletsprojects.com/en/2.1.x/api/#flask.copy_current_request_context
@copy_current_request_context
def resolveListenerCallWithinAContext(
    args,
    kwargs,
    wrapperManager,
    roleRequired,
    apiKeyRequired,
    contextRequired,
    requestHeaderClass,
    requestParamClass,
    requestClass,
    responseClass,
    defaultResponseHeaders,
    consumes,
    resourceInstanceMethodMuteStacktraceOnBusinessRuleException,

    requestUrl,
    requestVerb,
    requestHeaders,
    requestParams,
    requestBody,

    logRequestMessage = LogConstant.LISTENER_REQUEST,
    context = HttpDomain.LISTENER_CONTEXT
):
    ###- https://flask.palletsprojects.com/en/2.1.x/appcontext/
    ###- https://flask.palletsprojects.com/en/2.1.x/reqcontext/
    ###- https://werkzeug.palletsprojects.com/en/2.1.x/test/#werkzeug.test.EnvironBuilder
    # with wrapperManager.api.app.test_request_context(
    #     path = requestUrl,
    #     method = requestVerb,
    #     headers = requestHeaders,
    #     ###- query_string = requestParams, ###- query string already comes in the url
    #     json = requestBody
    # ):
    ###- https://flask.palletsprojects.com/en/1.1.x/appcontext/
    # with wrapperManager.api.app.app_context():
    resolveListenerCall(
        args,
        kwargs,
        wrapperManager,
        roleRequired,
        apiKeyRequired,
        contextRequired,
        requestHeaderClass,
        requestParamClass,
        requestClass,
        responseClass,
        defaultResponseHeaders,
        consumes,
        resourceInstanceMethodMuteStacktraceOnBusinessRuleException,

        requestUrl,
        requestVerb,
        requestHeaders,
        requestParams,
        requestBody,

        logRequestMessage = LogConstant.LISTENER_REQUEST,
        context = HttpDomain.LISTENER_CONTEXT
    )



def resolveListenerCall(
    args,
    kwargs,
    wrapperManager,
    roleRequired,
    apiKeyRequired,
    contextRequired,
    requestHeaderClass,
    requestParamClass,
    requestClass,
    responseClass,
    defaultResponseHeaders,
    consumes,
    resourceInstanceMethodMuteStacktraceOnBusinessRuleException,

    requestUrl,
    requestVerb,
    requestHeaders,
    requestParams,
    requestBody,

    logRequestMessage = LogConstant.LISTENER_REQUEST,
    context = HttpDomain.LISTENER_CONTEXT
):
    completeResponse = None
    try:
        completeResponse = FlaskManager.handleAnyControllerMethodRequest(
            args,
            kwargs,
            consumes,
            wrapperManager.resourceInstance,
            wrapperManager.resourceInstanceMethod,
            contextRequired,
            apiKeyRequired,
            roleRequired,
            requestHeaderClass,
            requestParamClass,
            requestClass,
            wrapperManager.shouldLogRequest(),
            resourceInstanceMethodMuteStacktraceOnBusinessRuleException,
            verb = requestVerb,
            requestHeaders = requestHeaders,
            requestParams = requestParams,
            requestBody = requestBody,
            logRequestMessage = logRequestMessage
        )
        FlaskManager.validateCompleteResponse(responseClass, completeResponse)
    except Exception as exception:
        log.log(resolveListenerCall, 'Failure at controller method execution. Getting complete response as exception', exception=exception, muteStackTrace=True)
        completeResponse = FlaskManager.getCompleteResponseByException(
            exception,
            wrapperManager.resourceInstance,
            wrapperManager.resourceInstanceMethod,
            resourceInstanceMethodMuteStacktraceOnBusinessRuleException,
            context = context
        )
    try:
        status = HttpStatus.map(completeResponse[-1])
        additionalResponseHeaders = completeResponse[1]
        if ObjectHelper.isNotNone(wrapperManager.resourceInstance.responseHeaders):
            additionalResponseHeaders = {**wrapperManager.resourceInstance.responseHeaders, **additionalResponseHeaders}
        if ObjectHelper.isNotNone(defaultResponseHeaders):
            additionalResponseHeaders = {**defaultResponseHeaders, **additionalResponseHeaders}
        responseBody = completeResponse[0] if ObjectHelper.isNotNone(completeResponse[0]) else {'message' : status.enumName}
        completeResponse = [responseBody, additionalResponseHeaders, status]
    except Exception as exception:
        log.failure(resolveListenerCall, f'Failure while parsing complete response: {completeResponse}. Returning simplified version of it', exception, muteStackTrace=True)
        completeResponse = getCompleteResponseByException(
            Exception(f'Not possible to handle complete response{c.DOT_SPACE_CAUSE}{str(exception)}'),
            wrapperManager.resourceInstance,
            wrapperManager.resourceInstanceMethod,
            resourceInstanceMethodMuteStacktraceOnBusinessRuleException
        )
    return completeResponse



from python_helper import Constant as c
from python_helper import ObjectHelper, StringHelper
from python_framework import ConverterStatic
from python_framework import SqlAlchemyProxy as sap

from ModelAssociation import EMISSION, MESSAGE, MODEL
from util import ModelUtil
from constant import EmissionConstant, ModelConstant
from helper.static import EmissionModelHelperStatic


class EmissionModel(MODEL):
    __tablename__ = EMISSION

    id = sap.Column(sap.Integer(), sap.Sequence(f'{__tablename__}{sap.ID}{sap.SEQ}'), primary_key=True)
    key = sap.Column(sap.String(3*sap.MEDIUM_STRING_SIZE), nullable=False, unique=True)
    queueKey = sap.Column(sap.String(sap.MEDIUM_STRING_SIZE), nullable=False)
    subscriptionKey = sap.Column(sap.String(sap.MEDIUM_STRING_SIZE), nullable=False)
    groupKey = sap.Column(sap.String(sap.MEDIUM_STRING_SIZE), nullable=False)
    messageKey = sap.Column(sap.String(sap.MEDIUM_STRING_SIZE), nullable=False)
    originKey = sap.Column(sap.String(sap.MEDIUM_STRING_SIZE), nullable=False)

    url = sap.Column(sap.String(sap.LARGE_STRING_SIZE), nullable=False)
    tries = sap.Column(sap.Integer(), nullable=False, default=EmissionConstant.ZERO_TRIES)
    onErrorUrl = sap.Column(sap.String(sap.MEDIUM_STRING_SIZE))
    onErrorTries = sap.Column(sap.Integer(), nullable=False, default=EmissionConstant.ZERO_TRIES)
    maxTries = sap.Column(sap.Integer(), nullable=False, default=EmissionConstant.DEFAULT_MAX_TRIES)
    backOff = sap.Column(sap.Float(precision=3), nullable=False, default=EmissionConstant.DEFAULT_BACKOFF)
    status = sap.Column(sap.String(sap.LITTLE_STRING_SIZE), nullable=False, default=ModelConstant.DEFAULT_STATUS)
    state = sap.Column(sap.String(sap.LITTLE_STRING_SIZE), nullable=False, default=ModelConstant.DEFAULT_STATE)

    history = sap.Column(sap.String(65_536))

    createdAt = sap.Column(sap.DateTime, nullable=False)
    updatedAt = sap.Column(sap.DateTime, nullable=False)


    def __init__(self,
        id = None,
        key = None,
        queueKey = None,
        subscriptionKey = None,
        groupKey = None,
        messageKey = None,
        originKey = None,
        url = None,
        tries = None,
        onErrorUrl = None,
        onErrorTries = None,
        maxTries = None,
        backOff = None,
        status = None,
        state = None,
        history = None,
        createdAt = None,
        updatedAt = None
    ):
        self.id = id
        self.queueKey = queueKey
        self.subscriptionKey = subscriptionKey
        self.groupKey = groupKey
        self.messageKey = messageKey
        self.originKey = originKey
        self.url = url
        self.tries = ConverterStatic.getValueOrDefault(tries, EmissionConstant.ZERO_TRIES)
        self.onErrorUrl = onErrorUrl
        self.onErrorTries = ConverterStatic.getValueOrDefault(onErrorTries, EmissionConstant.ZERO_TRIES)
        self.maxTries = ConverterStatic.getValueOrDefault(maxTries, EmissionConstant.DEFAULT_MAX_TRIES)
        self.backOff = ConverterStatic.getValueOrDefault(backOff, EmissionConstant.DEFAULT_BACKOFF)
        self.status = ConverterStatic.getValueOrDefault(status, ModelConstant.DEFAULT_STATUS)
        self.state = ConverterStatic.getValueOrDefault(state, ModelConstant.DEFAULT_STATE)
        self.setHistory(history)
        self.updateKey(key=key)

        self.createdAt = createdAt
        self.updatedAt = updatedAt
        ConverterStatic.overrideDateData(self)


    def updateKey(self, key=None):
        self.key = ConverterStatic.getValueOrDefault(key, EmissionModelHelperStatic.buildKey(
            queueKey = self.queueKey,
            subscriptionKey = self.subscriptionKey,
            messageKey = self.messageKey
        ))


    def setHistory(self, history):
        self.history = str(history)


    def addHistory(self, history):
        self.history = str(history)


    def __repr__(self):
        return f'{self.__tablename__}(id={self.id}, key={self.key}, queueKey={self.queueKey}, subscriptionKey={self.subscriptionKey}, messageKey={self.messageKey}, tries={self.tries}, onErrorTries={self.onErrorTries}, maxTries={self.maxTries}, backOff={self.backOff}, status={self.status}, state={self.state})'



from python_helper import ObjectHelper
from python_framework import Validator, ValidatorMethod, GlobalException, HttpStatus

import QueueDto
from SubscriptionModel import SubscriptionModel

@Validator()
class QueueModelValidator:

    @ValidatorMethod(requestClass=[QueueDto.QueueRequestDto])
    def validateRequestDto(self, dto):
        if ObjectHelper.isNone(dto) or ObjectHelper.isNone(dto.key):
            raise GlobalException(
                message = f'Queue key cannot be None',
                logMessage = f'Subscription: {dto}',
                status = HttpStatus.BAD_REQUEST
            )


    @ValidatorMethod(requestClass=[QueueDto.QueueRequestDto])
    def validateDoesNotExists(self, dto):
        self.validateRequestDto(dto)
        if self.service.queueModel.existsByKey(dto.key):
            raise GlobalException(
                message = f'Queue aleady exists',
                logMessage = f'Queue {dto.key} aleady exists',
                status = HttpStatus.BAD_REQUEST
            )


    @ValidatorMethod(requestClass=[str])
    def validateExistsByKey(self, key):
        if not self.service.queueModel.existsByKey(key):
            raise GlobalException(
                message = 'Queue does not exists',
                logMessage = f'Queue {key} does not exists',
                status = HttpStatus.BAD_REQUEST
            )



from python_framework import ConverterStatic

try:
    import SubscriptionConstant
except:
    from queue_manager_api.api.src.constant import SubscriptionConstant


class SubscriptionRequestDto:
    def __init__(self,
        key = None,
        url = None,
        onErrorUrl = None,
        maxTries = None,
        backOff = None,
        queue = None
    ):
        self.key = key
        self.url = url
        self.onErrorUrl = onErrorUrl
        self.maxTries = ConverterStatic.getValueOrDefault(maxTries, SubscriptionConstant.DEFAULT_MAX_TRIES)
        self.backOff = ConverterStatic.getValueOrDefault(backOff, SubscriptionConstant.DEFAULT_BACKOFF)
        self.queue = queue


class SubscriptionResponseDto:
    def __init__(self,
        key = None,
        url = None,
        onErrorUrl = None,
        maxTries = None,
        backOff = None,
        headers = None,
        queue = None
    ):
        self.key = key
        self.url = url
        self.onErrorUrl = onErrorUrl
        self.maxTries = ConverterStatic.getValueOrDefault(maxTries, SubscriptionConstant.DEFAULT_MAX_TRIES)
        self.backOff = ConverterStatic.getValueOrDefault(backOff, SubscriptionConstant.DEFAULT_BACKOFF)
        self.queue = queue



from python_helper import Constant as c
from python_helper import ObjectHelper


def containsEmission(emissionList, emission):
    return (
        ObjectHelper.isNotNone(emissionList) and
        ObjectHelper.isNotNone(emission) and
        getKey(emission) in [getKey(e) for e in emissionList]
    )


def getKey(emission):
    return emission.key if ObjectHelper.isNotNone(emission.key) else buildKey(
        queueKey = emission.queueKey,
        subscriptionKey = emission.subscriptionKey,
        messageKey = emission.messageKey
    )


def buildKey(queueKey=None, subscriptionKey=None, messageKey=None):
    return f'{queueKey}{c.DOT}{subscriptionKey}{c.DOT}{messageKey}'



from python_framework import Controller, ControllerMethod, HttpStatus

from enumeration.AccessDomain import AccessDomain
import MessageDto


@Controller(url = '/message/emitter', tag='Message', description='Message controller')
class MessageController:

    @ControllerMethod(url = '/',
        apiKeyRequired = [AccessDomain.API, AccessDomain.USER],
        requestClass = [MessageDto.MessageRequestDto],
        responseClass = [MessageDto.MessageCreationResponseDto]
        , logRequest = True
        , logResponse = True
    )
    def post(self, dto):
        return self.service.message.accept(dto), HttpStatus.ACCEPTED



from python_helper import log, ObjectHelper
from python_framework import Service, ServiceMethod

import Message, Emission, QueueModel
from enumeration.ModelState import ModelState
from enumeration.ModelStatus import ModelStatus


LOG_LEVEL = log.STATUS


@Service()
class MemoryService:

    @ServiceMethod()
    def findAll(self):
        return self.repository.memory.findAll()


    @ServiceMethod(requestClass=[Message.Message])
    def acceptMessage(self, message):
        self.validator.memory.validateMessageDoesNotExists(message)
        log.prettyPython(self.acceptMessage, f'Accepting new queued message', message, logLevel=LOG_LEVEL)
        return self.repository.memory.acceptMessage(message)


    @ServiceMethod(requestClass=[[Emission.Emission], QueueModel.QueueModel])
    def acceptEmissionList(self, emissionList, queue):
        self.validator.memory.validateThereAreAtLeastOneEmission(emissionList, queue)
        log.prettyPython(self.acceptEmissionList, f'Accepting queued message emissions', emissionList, logLevel=LOG_LEVEL)
        return self.repository.memory.acceptEmissionList(emissionList, queue)


    @ServiceMethod(requestClass=[Message.Message])
    def messageExists(self, message):
        return self.repository.memory.existsMessageByQueueKeyAndMessageKey(message.queueKey, message.key)


    @ServiceMethod()
    def getAllInstantiatedMessages(self):
        return self.repository.memory.findAllMessagesByStateIn([ModelState.INSTANTIATED])


    @ServiceMethod()
    def getAllInstantiatedEmissions(self):
        return self.repository.memory.findAllEmissionsByStateIn([ModelState.INSTANTIATED])


    @ServiceMethod()
    def getAllAcceptedMessages(self):
        return self.repository.memory.findAllMessagesByStatusIn([ModelStatus.ACCEPTED])


    @ServiceMethod()
    def getAllModifiedMessages(self):
        return self.repository.memory.findAllMessagesByStateIn([
            ModelState.INSTANTIATED,
            ModelState.MODIFIED
        ])


    @ServiceMethod()
    def getAllAcceptedMessagesFromOneQueue(self):
        return self.repository.memory.findAllMessagesByStatusInFromOneQueue([ModelStatus.ACCEPTED])


    @ServiceMethod()
    def getAcceptedEmissionsFromOneQueue(self):
        return self.repository.memory.findAllEmissionsByStatusInFromOneQueue([ModelStatus.ACCEPTED])


    @ServiceMethod()
    def getAllModifiedEmissions(self):
        return self.repository.memory.findAllEmissionsByStateIn([
            ModelState.INSTANTIATED,
            ModelState.MODIFIED
        ])


    @ServiceMethod()
    def removeAllProcessedMessagesAndEmissions(self):
        FINAL_STATE = [
            ModelState.PROCESSED
        ]
        FINAL_STATUS = [
            ModelStatus.PROCESSED,
            ModelStatus.PROCESSED_WITH_ERRORS,
            ModelStatus.UNPROCESSED
        ]
        messageList = self.repository.memory.removeAllMessagesByStateInAndSatusIn(FINAL_STATE, FINAL_STATUS)
        if ObjectHelper.isNotEmpty(messageList):
            self.service.messageModel.createOrUpdateAll(messageList)
            log.prettyPython(self.removeAllProcessedMessagesAndEmissions, 'Messages removed from memory', messageList, logLevel=LOG_LEVEL)
        emissionList = self.repository.memory.removeAllEmissionsByStateInAndSatusIn(FINAL_STATE, FINAL_STATUS)
        if ObjectHelper.isNotEmpty(emissionList):
            self.service.emissionModel.createOrUpdateAll(emissionList)
            log.prettyPython(self.removeAllProcessedMessagesAndEmissions, 'Emissions removed from memory', emissionList, logLevel=LOG_LEVEL)



from python_helper import Constant as c
from python_helper import ObjectHelper
from python_framework import ConverterStatic, Serializer
from python_framework import SqlAlchemyProxy as sap

from ModelAssociation import MESSAGE, EMISSION
from constant import ModelConstant


class Message:
    __memoryname__ = MESSAGE.replace(Serializer.MODEL_SUFIX, c.BLANK)

    def __init__(self,
        key = None,
        queueKey = None,
        groupKey = None,
        originKey = None,
        headers = None,
        content = None,
        status = None,
        state = None,
        emissionList = None,
        history = None
    ):
        self.key = key
        self.queueKey = queueKey
        self.groupKey = groupKey
        self.originKey = originKey
        self.headers = headers
        self.content = content
        self.status = ConverterStatic.getValueOrDefault(status, ModelConstant.DEFAULT_STATUS)
        self.state = ConverterStatic.getValueOrDefault(state, ModelConstant.DEFAULT_STATE)
        self.setEmissionList(emissionList)
        self.setHistory(history)


    def setEmissionList(self, emissionList):
        self.emissionList = ConverterStatic.getValueOrDefault(emissionList, [])


    def addEmissionList(self, emissionList):
        for emission in emissionList:
            self.addEmission(emission)


    def addEmission(self, emission):
        if not EmissionModelHelperStatic.containsEmission(self.emissionList, emission):
            self.emissionList.append(emission)


    def setHistory(self, history):
        if ObjectHelper.isNone(history):
            self.history = []
        elif ObjectHelper.isNotList(history):
            self.history = [str(history)]
        else:
            self.history = [
                str(h) for h in ConverterStatic.getValueOrDefault(history, [])
            ]
        # self.history = []


    def addHistory(self, history):
        self.history.append(str(history))
        # self.history = []


    def __repr__(self):
        return f'{self.__memoryname__}(key={self.key}, queueKey={self.queueKey}, groupKey={self.groupKey}, originKey={self.originKey}, status={self.status}, state={self.state})'



from python_framework import ConverterStatic


class QueueRequestDto:
    def __init__(self,
        key = None,
        subscriptionList = None
    ):
        self.key = key
        self.subscriptionList = ConverterStatic.getValueOrDefault(subscriptionList, [])


class QueueResponseDto:
    def __init__(self,
        key = None,
        subscriptionList = None
    ):
        self.key = key
        self.subscriptionList = ConverterStatic.getValueOrDefault(subscriptionList, [])



from python_helper import ObjectHelper
from python_framework import Mapper, MapperMethod

import QueueModel
import QueueDto


@Mapper()
class QueueModelMapper:

    @MapperMethod(requestClass=[[QueueDto.QueueRequestDto], str], responseClass=[[QueueModel.QueueModel]])
    def fromRequestDtoListToModelList(self, dtoList, originKey, modelList):
        for model in modelList:
            self.overrideModelOriginKey(originKey, model)
        return modelList


    @MapperMethod(requestClass=[[QueueModel.QueueModel]], responseClass=[[QueueDto.QueueResponseDto]])
    def fromModelListToResponseDtoList(self, modelList, dtoList):
        return dtoList


    @MapperMethod(requestClass=[QueueDto.QueueRequestDto, str], responseClass=[QueueModel.QueueModel])
    def fromRequestDtoToModel(self, dto, originKey, model):
        self.overrideModelOriginKey(originKey, model)
        return model


    @MapperMethod(requestClass=[QueueModel.QueueModel], responseClass=[QueueDto.QueueResponseDto])
    def fromModelToResponseDto(self, model, dto):
        return dto


    @MapperMethod(requestClass=[QueueModel.QueueModel, QueueDto.QueueRequestDto])
    def overrideModel(self, model, dto):
        subscriptionDictionary = {subscription.key: subscription for subscription in model.subscriptionList}
        for subscriptionDto in dto.subscriptionList:
            if subscriptionDto.key in subscriptionDictionary:
                self.mapper.subscriptionModel.overrideModel(
                    subscriptionDictionary.get(subscriptionDto.key),
                    subscriptionDto
                )
            else:
                model.subscriptionList.append(self.mapper.subscriptionModel.fromRequestDtoToModel(subscriptionDto, model.originKey))


    @MapperMethod(requestClass=[str, QueueModel.QueueModel])
    def overrideModelOriginKey(self, originKey, model):
        model.originKey = originKey
        self.mapper.subscriptionModel.overrideAllModelOriginKey(originKey, model.subscriptionList)



from python_helper import ObjectHelper
from python_framework import Validator, ValidatorMethod, GlobalException, HttpStatus

import EmissionDto
import Emission


@Validator()
class EmissionValidator:

    @ValidatorMethod(requestClass=[EmissionDto.EmissionRequestDto])
    def validateRequestDto(self, dto):
        self.validator.emissionModel.validateRequestDto(dto)


    @ValidatorMethod(requestClass=[EmissionDto.EmissionRequestDto])
    def validateDoesNotExists(self, dto):
        self.validator.emissionModel.validateRequestDto(dto)


    @ValidatorMethod(requestClass=[[Emission.Emission]])
    def validateAllBelongsToTheSameQueue(self, modelList):
        if not 1 == len({model.queueKey for model in modelList if ObjectHelper.isNotNone(model.queueKey)}):
            raise GlobalException(
                logEmission = f'All emissions should be from the same queue. Emissions: {modelList}',
                status = HttpStatus.INTERNAL_SERVER_ERROR
            )



from python_helper import Constant as c
from python_helper import ObjectHelper, StringHelper
from python_framework import SqlAlchemyProxy as sap
from python_framework import ConverterStatic

from ModelAssociation import SUBSCRIPTION, QUEUE, MODEL
from constant import SubscriptionConstant
from util import ModelUtil


class SubscriptionModel(MODEL):
    __tablename__ = SUBSCRIPTION

    id = sap.Column(sap.Integer(), sap.Sequence(f'{__tablename__}{sap.ID}{sap.SEQ}'), primary_key=True)
    key = sap.Column(sap.String(sap.MEDIUM_STRING_SIZE), nullable=False, unique=True)
    originKey = sap.Column(sap.String(sap.MEDIUM_STRING_SIZE), nullable=False)
    url = sap.Column(sap.String(sap.LARGE_STRING_SIZE), nullable=False)
    onErrorUrl = sap.Column(sap.String(sap.MEDIUM_STRING_SIZE))
    maxTries = sap.Column(sap.Integer(), nullable=False, default=SubscriptionConstant.DEFAULT_MAX_TRIES)
    backOff = sap.Column(sap.Float(precision=3), nullable=False, default=SubscriptionConstant.DEFAULT_BACKOFF)

    queue, queueId = sap.getManyToOne(SUBSCRIPTION, QUEUE, MODEL)

    def __init__(self,
        id = None,
        key = None,
        originKey = None,
        url = None,
        onErrorUrl = None,
        maxTries = None,
        backOff = None,
        headers = None,
        queue = None,
        queueId = None
    ):
        self.id = id
        self.key = key
        self.originKey = originKey
        self.url = url
        self.onErrorUrl = onErrorUrl
        self.maxTries = ConverterStatic.getValueOrDefault(maxTries, SubscriptionConstant.DEFAULT_MAX_TRIES)
        self.backOff = ConverterStatic.getValueOrDefault(backOff, SubscriptionConstant.DEFAULT_BACKOFF)
        self.setQueue(queue, queueId=queueId)


    def getQueueKey(self):
        return None if ObjectHelper.isNone(self.queue) else self.queue.key


    def setQueue(self, queue, queueId=None):
        self.queue, self.queueId = ModelUtil.getManyToOneData(queue, queueId, MODEL)


    def __repr__(self):
        return f'{self.__tablename__}(id={self.id}, key={self.key}, queueKey={self.getQueueKey()}, originKey={self.originKey}, maxTries={self.maxTries}, backOff={self.backOff}, url={self.url}, onErrorUrl={self.onErrorUrl})'



from python_helper import Constant as c
from python_helper import ObjectHelper, StringHelper
from python_framework import Helper, MapperMethod

import EmissionModel
import EmissionModelHelperStatic


@Helper()
class EmissionModelHelper:

    @MapperMethod(requestClass=[[EmissionModel.EmissionModel]])
    def getModelKeyList(self, modelList):
        return [self.getModelKey(model) for model in modelList]


    @MapperMethod(requestClass=[EmissionModel.EmissionModel])
    def getModelKey(self, model):
        return EmissionModelHelperStatic.getKey(model)



MESSAGE_KEY_CLIENT_ATTRIBUTE_NAME = 'messageKey'
GROUP_KEY_CLIENT_ATTRIBUTE_NAME = 'groupKey'
MESSAGE_HEADERS_KEY_CLIENT_ATTRIBUTE_NAME = 'messageHeaders'

MESSAGE_KEY_KEY = 'key'
MESSAGE_QUEUE_KEY_KEY = 'queueKey'
MESSAGE_GROUP_KEY = 'groupKey'
MESSAGE_CONTENT_KEY = 'content'



from python_framework import Enum, EnumItem


@Enum()
class ModelStateEnumeration :
    INSTANTIATED = EnumItem()
    PERSISTED = EnumItem()
    MODIFIED = EnumItem()
    PROCESSED = EnumItem()
    NOT_INFORMED = EnumItem()


ModelState = ModelStateEnumeration()



from python_helper import ObjectHelper


def getManyToOneData(model, modelId, refferenceModel):
    return model, modelId if ObjectHelper.isNone(model) or not isinstance(model, refferenceModel.__class__) else model.id


def getOneToManyData(modelList):
    return [] if ObjectHelper.isEmpty(modelList) else modelList



import functools
import operator
from python_helper import ObjectHelper
from python_framework import Repository

from enumeration.ModelState import ModelState


MESSAGE_MODEL = 'message'
EMISSION_MODEL = 'emission'


@Repository()
class MemoryRepository:

    data = {}


    def existsMessageByQueueKeyAndMessageKey(self, queueKey, messageKey):
        return messageKey in self.getMessageQueue(queueKey)


    def findAllMessagesByStateIn(self, stateList):
        return [
            self.getMessage(queueKey, messageKey)
            for queueKey in self.getQueueKeyIterator()
            for messageKey in self.getMessageKeyIterator(queueKey)
            if self.messageStateIn(queueKey, messageKey, stateList)
        ]


    def removeAllMessagesByStateIn(self, stateList):
        messageList = [
            self.popMessage(queueKey, messageKey)
            for queueKey in self.getQueueKeyIterator()
            for messageKey in self.getMessageKeyIterator(queueKey)
            if self.messageStateIn(queueKey, messageKey, stateList)
        ]
        self.removeEmptyQueues()
        return messageList


    def removeAllMessagesByStateInAndSatusIn(self, stateList, statusList):
        messageList = [
            self.popMessage(queueKey, messageKey)
            for queueKey in self.getQueueKeyIterator()
            for messageKey in self.getMessageKeyIterator(queueKey)
            if self.messageStateIn(queueKey, messageKey, stateList) and self.messageStatusIn(queueKey, messageKey, statusList)
        ]
        self.removeEmptyQueues()
        return messageList


    def findAllMessagesByStatusIn(self, statusList):
        return [
            self.getMessage(queueKey, messageKey)
            for queueKey in self.getQueueKeyIterator()
            for messageKey in self.getMessageKeyIterator(queueKey)
            if self.messageStatusIn(queueKey, messageKey, statusList)
        ]


    def removeAllMessageByStatusIn(self, statusList):
        messageList = [
            self.popMessage(queueKey, messageKey)
            for queueKey in self.getQueueKeyIterator()
            for messageKey in self.getMessageKeyIterator(queueKey)
            if self.messageStatusIn(queueKey, messageKey, statusList)
        ]
        self.removeEmptyQueues()
        return messageList


    def findAllMessagesByStatusInFromOneQueue(self, statusList):
        queueKeyList = self.getQueueKeyIterator()
        if ObjectHelper.isEmpty(queueKeyList):
            return []
        for queueKey in self.getQueueKeyIterator():
            messageList = [
                self.getMessage(queueKey, messageKey)
                for messageKey in self.getMessageKeyIterator(queueKey)
                if self.messageStatusIn(queueKey, messageKey, statusList)
            ]
            if 0 < len(messageList):
                return messageList
        return []


    def getMessageKeyIterator(self, queueKey):
        return [*self.getMessageQueue(queueKey).keys()]


    def getEmissionMessageKeyIterator(self, queueKey):
        return [*self.getEmissionQueue(queueKey).keys()]


    def getMessage(self, queueKey, messageKey):
        return self.getMessageQueue(queueKey).get(messageKey)


    def popMessage(self, queueKey, messageKey):
        return self.getMessageQueue(queueKey).pop(messageKey)


    def getMessageQueue(self, queueKey):
        return self.getQueue(queueKey).get(MESSAGE_MODEL, {})


    def getEmissionQueue(self, queueKey):
        return self.getQueue(queueKey).get(EMISSION_MODEL, {})


    def getEmissionList(self, queueKey, messageKey):
        return self.getEmissionQueue(queueKey).get(messageKey, [])


    def findAllEmissionsByStateIn(self, stateList):
        return [
            emission
            for queueKey in self.getQueueKeyIterator()
            for messageKey in self.getEmissionMessageKeyIterator(queueKey)
            for emission in self.getEmissionList(queueKey, messageKey)
            if self.emissionStateIn(emission, stateList)
        ]


    def findAllEmissionsByStatusInFromOneQueue(self, statusList):
        queueKeyList = self.getQueueKeyIterator()
        if ObjectHelper.isEmpty(queueKeyList):
            return []
        for queueKey in queueKeyList:
            emissionList = [
                emission
                for messageKey in self.getMessageKeyIterator(queueKey)
                for emission in self.getEmissionList(queueKey, messageKey)
                if self.emissionStatusIn(emission, statusList)
            ]
            if 0 < len(emissionList):
                return emissionList
        return []


    def removeAllEmissionsByStateInAndSatusIn(self, stateList, statusList):
        emissionList = [
            self.popEmission(emission)
            for queueKey in self.getQueueKeyIterator()
            for messageKey in self.getEmissionMessageKeyIterator(queueKey)
            for emission in self.getEmissionList(queueKey, messageKey)
            if self.emissionStateIn(emission, stateList) and self.emissionStatusIn(emission, statusList)
        ]
        self.removeEmptyQueues()
        return emissionList


    def removeEmptyQueues(self):
        for queueKey in self.getQueueKeyIterator():
            if ObjectHelper.isEmpty(self.getMessageQueue(queueKey)) and ObjectHelper.isEmpty(self.getEmissionQueue(queueKey)):
                ObjectHelper.deleteDictionaryEntry(queueKey, self.data)


    def emissionStatusIn(self, emission, statusList):
        return ObjectHelper.isNotNone(emission) and emission.status in statusList


    def emissionStateIn(self, emission, stateList):
        return ObjectHelper.isNotNone(emission) and emission.state in stateList


    def popEmission(self, emission):
        emissionList = self.getEmissionQueue(emission.queueKey).get(emission.getMessageKey())
        emissionIndex = [
            i
            for i, e in enumerate(emissionList)
            if e == emission
        ]
        if ObjectHelper.isEmpty(emissionIndex):
            raise Exception(f'Emission {emission} not found')
        if 1 == len(emissionList):
            self.getEmissionQueue(emission.queueKey).pop(emission.getMessageKey())
        else:
            self.getEmissionQueue(emission.queueKey).get(emission.getMessageKey()).pop(emissionIndex[0])
        return emission


    def messageStateIn(self, queueKey, messageKey, stateList):
        return ObjectHelper.isNotNone(self.getMessage(queueKey, messageKey)) and self.getMessage(queueKey, messageKey).state in stateList


    def messageStatusIn(self, queueKey, messageKey, statusList):
        return ObjectHelper.isNotNone(self.getMessage(queueKey, messageKey)) and self.getMessage(queueKey, messageKey).status in statusList


    def acceptMessage(self, message):
        self.addQueueKeyIfNeeded(message.queueKey)
        self.data[message.queueKey][MESSAGE_MODEL][message.key] = message
        return message


    def acceptEmissionList(self, emissionList, queue):
        if ObjectHelper.isNotEmpty(emissionList):
            message = emissionList[0].message
            self.addQueueKeyIfNeeded(queue.key)
            self.data[queue.key][EMISSION_MODEL][message.key] = emissionList


    def addQueueKeyIfNeeded(self, queueKey):
        if queueKey not in self.data:
            self.data[queueKey] = {
                MESSAGE_MODEL: {},
                EMISSION_MODEL: {}
            }


    def getQueueKeyIterator(self):
        return [*{**self.data}.keys()]


    def getQueue(self, queueKey):
        return self.data.get(queueKey, {})


    def findAll(self):
        return {**self.data}



from python_framework import Controller, ControllerMethod, HttpStatus

from enumeration.AccessDomain import AccessDomain
import QueueDto


@Controller(url = '/queue', tag='Queue', description='Queue controller')
class QueueModelController:

    @ControllerMethod(url = '/',
        apiKeyRequired = [AccessDomain.API],
        requestClass = [QueueDto.QueueRequestDto],
        responseClass = [QueueDto.QueueResponseDto]
        # , logRequest = True
        # , logResponse = True
    )
    def post(self, dto):
        return self.service.queueModel.createOrUpdate(dto), HttpStatus.CREATED


@Controller(url = '/queue/all', tag='Queue', description='Queue controller')
class QueueModelAllController:

    @ControllerMethod(url = '/',
        apiKeyRequired = [AccessDomain.API],
        responseClass = [[QueueDto.QueueResponseDto]]
        # , logRequest = True
        # , logResponse = True
    )
    def get(self):
        return self.service.queueModel.findAllByOrigin(), HttpStatus.OK



import time
from python_helper import log, ObjectHelper
from python_framework import Service, ServiceMethod

from constant import EmissionConstant
from enumeration.ModelState import ModelState
from enumeration.ModelStatus import ModelStatus

import MessageDto
import Emission, Message, SubscriptionModel, QueueModel


LOG_LEVEL = log.STATUS


@Service()
class EmissionService:

    @ServiceMethod(requestClass=[Message.Message, SubscriptionModel.SubscriptionModel, QueueModel.QueueModel])
    def buildNewModel(self, message, subscription, queue):
        model = Emission.Emission(
            queueKey = queue.key,
            subscriptionKey = subscription.key,
            url = subscription.url,
            tries = EmissionConstant.ZERO_TRIES,
            onErrorUrl = subscription.onErrorUrl,
            onErrorTries = EmissionConstant.ZERO_TRIES,
            maxTries = subscription.maxTries,
            backOff = subscription.backOff,
            state = ModelState.INSTANTIATED,
            message = message
        )
        message.emissionList.append(model)
        return model


    @ServiceMethod(requestClass=[Message.Message, QueueModel.QueueModel])
    def buildNewModelList(self, message, queue):
        return [
            self.buildNewModel(
                message,
                subscription,
                queue
            ) for subscription in queue.subscriptionList
        ]


    @ServiceMethod(requestClass=[[Message.Message], QueueModel.QueueModel])
    def acceptWithoutValidation(self, messageList, queue):
        for message in messageList:
            modelList = self.service.emission.buildNewModelList(message, queue)
            self.mapper.emission.overrideAllModelStatus(modelList, ModelStatus.ACCEPTED)
            log.prettyPython(self.acceptWithoutValidation, f'Accepting new queued message emissions', modelList, logLevel=LOG_LEVEL)
            self.service.memory.acceptEmissionList(modelList, queue)


    @ServiceMethod()
    def sendAllAcceptedFromOneQueue(self):
        modelList = self.service.memory.getAcceptedEmissionsFromOneQueue()
        if ObjectHelper.isEmpty(modelList):
            return []
        self.mapper.emission.overrideAllModelStatus(modelList, ModelStatus.PROCESSING)
        log.prettyPython(self.sendAllAcceptedFromOneQueue, f'Processing queued message emissions', modelList, logLevel=LOG_LEVEL)
        self.validator.emission.validateAllBelongsToTheSameQueue(modelList)
        for model in modelList:
            self.sendInAThread(model)


    @ServiceMethod(requestClass=[Emission.Emission])
    def sendInAThread(self, model):
        self.globals.api.resource.manager.queue.runInAThread(self.send, model)


    @ServiceMethod(requestClass=[Emission.Emission])
    def send(self, model):
        log.prettyPython(self.send, f'Sending queued message emission', model, logLevel=LOG_LEVEL)
        try:
            self.mapper.emission.overrideModelState(model, ModelState.MODIFIED)
            messageRequestDto = MessageDto.MessageRequestDto(
                key = model.getMessageKey(),
                queueKey = model.queueKey,
                groupKey = model.getGroupKey(),
                content = model.getContent()
            )
            emissionResponse = None
            if model.maxTries > model.tries:
                model.tries += 1
                emissionResponse = self.sendToDestiny(model.url, model.getHeaders(), messageRequestDto)
                self.mapper.emission.overrideModelStatus(model, ModelStatus.PROCESSED)
            elif ObjectHelper.isNeitherNoneNorBlank(model.onErrorUrl):
                model.onErrorTries += 1
                emissionResponse = self.sendToDestiny(model.onErrorUrl, model.getHeaders(), messageRequestDto)
                self.mapper.emission.overrideModelStatus(model, ModelStatus.PROCESSED_WITH_ERRORS)
            return emissionResponse
        except Exception as exception:
            time.sleep(model.backOff)
            model.addHistory(exception)
            if (
                model.maxTries > model.tries and ObjectHelper.isNoneOrBlank(model.onErrorUrl)
            ) or (
                model.maxTries > model.onErrorTries and ObjectHelper.isNeitherNoneNorBlank(model.onErrorUrl)
            ):
                log.warning(self.send, f'Error while trying to emit queued message emission "{model}". Going for another attempt', exception=exception, muteStackTrace=True)
                return self.send(model)
            else:
                self.mapper.emission.overrideModelStatus(model, ModelStatus.UNPROCESSED)
                log.failure(self.send, f'Queued message emission "{model}" processed with errors', exception=exception, muteStackTrace=True)
                return exception


    @ServiceMethod(requestClass=[str, dict, MessageDto.MessageRequestDto])
    def sendToDestiny(self, url, headers, messageRequestDto):
        return self.client.emission.send(url, messageRequestDto, headers=headers)


    @ServiceMethod()
    def updateAllModifiedFromMemory(self):
        modelList = self.service.memory.getAllModifiedEmissions()
        self.service.emissionModel.createOrUpdateAll(modelList)



from python_helper import ObjectHelper
from python_framework import Service, ServiceMethod

import QueueDto
import QueueModel
from util import AuditoryUtil


@Service()
class QueueModelService:

    @ServiceMethod()
    def findAllByOrigin(self):
        modelList = self.repository.queueModel.findAllByOriginKey(AuditoryUtil.getApiKeyIdentity(service=self))
        return self.mapper.queueModel.fromModelListToResponseDtoList(
            modelList
        )


    @ServiceMethod(requestClass=[QueueDto.QueueRequestDto])
    def createOrUpdate(self, dto):
        self.validator.queueModel.validateRequestDto(dto)
        model = self.findOptionalModelByKey(dto.key)
        if ObjectHelper.isNone(model):
            model = self.mapper.queueModel.fromRequestDtoToModel(dto, AuditoryUtil.getApiKeyIdentity(service=self))
        else:
            self.mapper.queueModel.overrideModel(model, dto)
        return self.mapper.queueModel.fromModelToResponseDto(self.persist(model))


    @ServiceMethod(requestClass=[str])
    def findModelByKey(self, key):
        self.validator.queueModel.validateExistsByKey(key)
        return self.findOptionalModelByKey(key)


    @ServiceMethod(requestClass=[str])
    def findOptionalModelByKey(self, key):
        return self.repository.queueModel.findByKey(key)


    @ServiceMethod(requestClass=[str])
    def existsByKey(self, key):
        return self.repository.queueModel.existsByKey(key)


    @ServiceMethod(requestClass=[QueueModel.QueueModel])
    def persist(self, model):
        return self.repository.queueModel.save(model)



from python_helper import ObjectHelper
from python_framework import Validator, ValidatorMethod, GlobalException, HttpStatus

import EmissionDto


@Validator()
class EmissionModelValidator:

    @ValidatorMethod(requestClass=[EmissionDto.EmissionRequestDto])
    def validateRequestDto(self, dto):
        if ObjectHelper.isNone(dto) or ObjectHelper.isNone(dto.key):
            raise GlobalException(
                message = f'Emission key cannot be None',
                logMessage = f'Subscription: {dto}',
                status = HttpStatus.BAD_REQUEST
            )


    @ValidatorMethod(requestClass=[EmissionDto.EmissionRequestDto])
    def validateDoesNotExists(self, dto):
        self.validateRequestDto(dto)
        self.validateDoesNotExistsByKey(dto.key)


    @ValidatorMethod(requestClass=[str])
    def validateDoesNotExistsByKey(self, key):
        if ObjectHelper.isNone(key) or self.service.emissionModel.existsByKey(key):
            raise GlobalException(
                message = 'Emission aleady exists',
                logMessage = f'Emission {key} aleady exists',
                status = HttpStatus.BAD_REQUEST
            )



from python_framework import HttpClientConstant, HttpDomain

CLIENT_DID_NOT_SENT_ANY_MESSAGE = HttpClientConstant.CLIENT_DID_NOT_SENT_ANY_MESSAGE.replace(HttpDomain.CLIENT_CONTEXT, 'Queue')
ERROR_AT_CLIENT_CALL_MESSAGE = HttpClientConstant.ERROR_AT_CLIENT_CALL_MESSAGE.replace(HttpDomain.CLIENT_CONTEXT.lower(), HttpDomain.EMITTER_CONTEXT.lower())



from python_framework import Controller, ControllerMethod, HttpStatus, ConverterStatic

from enumeration.AccessDomain import AccessDomain
import MessageDto


@Controller(url = '/test/message', tag='QueueTest', description='Queue test controller')
class MessateTestController:


    @ControllerMethod(url = '/',
        apiKeyRequired = ['SOMETHING_ELSE'],
        requestClass = [MessageDto.MessageRequestDto],
        responseClass = [MessageDto.MessageResponseDto]
        , logRequest = True
        , logResponse = True
    )
    def post(self, dto):
        return ConverterStatic.to(dto, MessageDto.MessageResponseDto), HttpStatus.OK


@Controller(url = '/test/message-error', tag='QueueTest', description='Queue test controller')
class MessateTestBulkController:

    @ControllerMethod(url = '/',
        apiKeyRequired = [AccessDomain.API],
        requestClass = [MessageDto.MessageRequestDto],
        responseClass = [MessageDto.MessageResponseDto]
        , logRequest = True
        , logResponse = True
    )
    def post(self, dto):
        return ConverterStatic.to(dto, MessageDto.MessageResponseDto), HttpStatus.OK



from python_helper import ObjectHelper
from python_framework import Validator, ValidatorMethod, GlobalException, HttpStatus

import Message, Emission, QueueModel


@Validator()
class MemoryValidator:

    @ValidatorMethod(requestClass=[Message.Message])
    def validateMessage(self, message):
        if ObjectHelper.isNone(message) or ObjectHelper.isNone(message.key) or ObjectHelper.isNone(message.queueKey):
            raise GlobalException(
                logMessage = f'Message key cannot be None. Message: {message}',
                status = HttpStatus.INTERNAL_SERVER_ERROR
            )


    @ValidatorMethod(requestClass=[Message.Message])
    def validateMessageDoesNotExists(self, message):
        self.validateMessage(message)
        if self.service.memory.messageExists(message):
            raise GlobalException(
                message = 'Message aleady in process',
                logMessage = f'Message "{message.key}" from subscription "{message.queueKey}" aleady in process',
                status = HttpStatus.BAD_REQUEST
            )


    @ValidatorMethod(requestClass=[[Emission.Emission], QueueModel.QueueModel])
    def validateThereAreAtLeastOneEmission(self, emissionList, queue):
        if ObjectHelper.isEmpty(emissionList):
            raise GlobalException(
                message = 'There are no subscriptions for this queue',
                logMessage = f'There are no subscriptions for the "{queue.key}" queue',
                status = HttpStatus.BAD_REQUEST
            )



from python_framework import ResourceManager

import ModelAssociation
from QueueManager import QueueManager


app = ResourceManager.initialize(__name__, ModelAssociation.MODEL, managerList=[
    QueueManager()
])



import requests, time
from flask import copy_current_request_context

from python_helper import Constant as c
from python_helper import ReflectionHelper, ObjectHelper, log, Function, StringHelper
from python_framework import (
    FlaskManager,
    ConverterStatic,
    FlaskUtil,
    ClientUtil,
    LogConstant,
    OpenApiManager,
    ConfigurationKeyConstant,
    HttpDomain,
    Serializer,
    GlobalException,
    HttpStatus
)

try:
    import AnnotationUtil
    import HttpEmitterConstant, MessageConstant, EmitterConstant
    import MessageDto
except:
    from queue_manager_api.api.src.util import AnnotationUtil
    from queue_manager_api.api.src.dto import MessageDto
    from queue_manager_api.api.src.constant import EmitterConstant
    from queue_manager_api.api.src.constant import HttpEmitterConstant
    from queue_manager_api.api.src.constant import MessageConstant


DEFAULT_TIMEOUT = 2


@Function
def MessageEmitter(
    *resourceArgs,
    url = EmitterConstant.URL,
    headers = None,
    timeout = DEFAULT_TIMEOUT,
    logRequest = False,
    logResponse = False,
    enabled = None,
    muteLogs = None,
    eventContext = HttpDomain.EMITTER_CONTEXT,
    **resourceKwargs
):
    def Wrapper(OuterClass, *outterArgs, **outterKwargs):
        resourceUrl = url
        resourceHeaders = ConverterStatic.getValueOrDefault(headers, dict())
        resourceTimeout = timeout
        resourceEventContext = eventContext
        resourceLogRequest = logRequest
        resourceLogResponse = logResponse
        resourceEnabled = enabled
        resourceMuteLogs = muteLogs
        log.wrapper(MessageEmitter, f'''wrapping {OuterClass.__name__}(*{outterArgs}, **{outterKwargs})''')
        api = FlaskManager.getApi()
        class InnerClass(OuterClass):
            url = resourceUrl
            headers = resourceHeaders
            def __init__(self, *args, **kwargs):
                log.wrapper(OuterClass, f'in {InnerClass.__name__}.__init__(*{args},**{kwargs})')
                OuterClass.__init__(self, *args,**kwargs)
                AnnotationUtil.initializeComunicationLayerResource(
                    resourceInstance = self,
                    api = api,
                    timeout = resourceTimeout,
                    enabled = resourceEnabled,
                    muteLogs = resourceMuteLogs,
                    logRequest = resourceLogRequest,
                    logResponse = resourceLogResponse,
                    resourceEnabledConfigKey = ConfigurationKeyConstant.API_EMITTER_ENABLE,
                    resourceMuteLogsConfigKey = ConfigurationKeyConstant.API_EMITTER_MUTE_LOGS,
                    resourceTimeoutConfigKey = ConfigurationKeyConstant.API_EMITTER_TIMEOUT
                )
                self.manager = api.resource.manager.queue
            def emit(self, *args, **kwargs):
                raise ClientUtil.HttpClientEvent(HttpDomain.Verb.POST, *args, eventContext=resourceEventContext, **kwargs)
        ReflectionHelper.overrideSignatures(InnerClass, OuterClass)
        return InnerClass
    return Wrapper


class EmitterMethodConfig:
    def __init__(self,
        url = None,
        headers = None,
        requestHeaderClass = None,
        requestParamClass = None,
        requestClass = None,
        responseClass = None,
        returnOnlyBody = None,
        timeout = DEFAULT_TIMEOUT,
        propagateAuthorization = None,
        propagateApiKey = None,
        propagateSession = None,
        produces = None,
        consumes = None,
        logRequest = None,
        logResponse = None,
        enabled = None,
        muteLogs = None
    ):
        self.url = url
        self.headers = headers
        self.requestHeaderClass = requestHeaderClass
        self.requestParamClass = requestParamClass
        self.requestClass = requestClass
        self.responseClass = responseClass
        self.returnOnlyBody = returnOnlyBody
        self.timeout = timeout
        self.propagateAuthorization = propagateAuthorization
        self.propagateApiKey = propagateApiKey
        self.propagateSession = propagateSession
        self.produces = produces
        self.consumes = consumes
        self.logRequest = logRequest
        self.logResponse = logResponse
        self.enabled = enabled
        self.muteLogs = muteLogs


@Function
def MessageEmitterMethod(
    *methodArgs,
    url = EmitterConstant.URL,
    headers = None,
    requestHeaderClass = None,
    requestParamClass = None,
    requestClass = None,
    responseClass = None,
    returnOnlyBody = True,
    timeout = None,
    consumes = OpenApiManager.DEFAULT_CONTENT_TYPE,
    produces = OpenApiManager.DEFAULT_CONTENT_TYPE,
    logRequest = True,
    logResponse = True,
    enabled = True,
    muteLogs = False,
    debugIt = False,
    muteStacktraceOnBusinessRuleException = True,
    queueKey = None,
    messageHeaders = None,
    **methodKwargs
):
    resourceMethodConfig = EmitterMethodConfig(
        url = url,
        headers = headers,
        requestHeaderClass = requestHeaderClass,
        requestParamClass = requestParamClass,
        requestClass = requestClass,
        responseClass = responseClass,
        returnOnlyBody = returnOnlyBody,
        timeout = timeout,
        propagateAuthorization = False,
        propagateApiKey = False,
        propagateSession = False,
        produces = produces,
        consumes = consumes,
        logRequest = logRequest,
        logResponse = logResponse,
        enabled = enabled,
        muteLogs = muteLogs
    )
    def innerMethodWrapper(resourceInstanceMethod, *innerMethodArgs, **innerMethodKwargs) :
        wrapperManager = AnnotationUtil.InnerMethodWrapperManager(
            wrapperType = MessageEmitterMethod,
            resourceInstanceMethod = resourceInstanceMethod,
            timeout = timeout,
            enabled = enabled,
            muteLogs = muteLogs,
            logRequest = logRequest,
            logResponse = logResponse,
            resourceTypeName = FlaskManager.KW_EMITTER_RESOURCE,
            resourceEnabledConfigKey = ConfigurationKeyConstant.API_EMITTER_ENABLE,
            resourceMuteLogsConfigKey = ConfigurationKeyConstant.API_EMITTER_MUTE_LOGS,
            resourceTimeoutConfigKey = ConfigurationKeyConstant.API_EMITTER_TIMEOUT
        )
        resourceInstanceMethodMuteStacktraceOnBusinessRuleException = muteStacktraceOnBusinessRuleException
        resourceMethodMessageHeaders = messageHeaders
        resourceMethodQueueKey = queueKey
        resourceMethodConfig.wrapperManager = wrapperManager
        def post(
            resourceInstance,
            body = None,
            url = None,
            headers = None,
            params = None,
            timeout = None,
            logRequest = False,
            messageKey = None,
            queueKey = resourceMethodQueueKey,
            groupKey = None,
            messageHeaders = None,
            **kwargs
        ):
            verb = HttpDomain.Verb.POST
            url, params, headers, body, timeout, logRequest = ClientUtil.parseParameters(
                resourceInstance,
                resourceMethodConfig,
                url,
                params,
                headers,
                Serializer.getObjectAsDictionary(
                    MessageDto.MessageRequestDto(
                        key = messageKey,
                        queueKey = queueKey,
                        groupKey = groupKey,
                        headers = {
                            **ConverterStatic.getValueOrDefault(resourceMethodMessageHeaders, dict()),
                            **ConverterStatic.getValueOrDefault(messageHeaders, dict())
                        },
                        content = body
                    )
                ),
                timeout,
                resourceMethodConfig.wrapperManager.shouldLogRequest()
            )
            doLogRequest(verb, url, body, params, headers, logRequest, kwargs)
            emitterMethodResponse = requests.post(
                url,
                params = params,
                headers = headers,
                json = body,
                timeout = timeout,
                **kwargs
            )
            return emitterMethodResponse

        def doLogRequest(verb, url, body, params, headers, logRequest, requestKwargs):
            log.info(resourceInstanceMethod, f'{LogConstant.EMITTER_SPACE}{verb}{c.SPACE_DASH_SPACE}{url}')
            if logRequest:
                parsetRequestKwargs = {} if ObjectHelper.isEmpty(requestKwargs) else {'requestKwargs': {**requestKwargs}}
                log.prettyJson(
                    resourceInstanceMethod,
                    LogConstant.EMITTER_REQUEST,
                    {
                        'headers': ConverterStatic.getValueOrDefault(headers, dict()),
                        'query': ConverterStatic.getValueOrDefault(params, dict()),
                        'body': ConverterStatic.getValueOrDefault(body, dict()),
                        **parsetRequestKwargs
                    },
                    condition = True,
                    logLevel = log.INFO
                )

        HTTP_CLIENT_RESOLVERS_MAP = {
            HttpDomain.Verb.POST : post
        }
        def innerResourceInstanceMethod(*args, **kwargs):
            f'''(*args, {FlaskUtil.KW_HEADERS}={{}}, {FlaskUtil.KW_PARAMETERS}={{}}, **kwargs)'''
            wrapperManager.updateResourceInstance(args)
            resourceMethodConfig.logRequest = wrapperManager.shouldLogRequest()
            httpClientResolversMap = HTTP_CLIENT_RESOLVERS_MAP
            messageCreationRequestKey = kwargs.get(MessageConstant.MESSAGE_KEY_CLIENT_ATTRIBUTE_NAME)
            if ObjectHelper.isNone(messageCreationRequestKey):
                messageCreationRequestKey = f'{f"{time.time():0<18}".replace(c.DOT, c.DASH)}{c.DASH}{Serializer.newUuid()}'
            messageCreationRequestGroupKey = ConverterStatic.getValueOrDefault(
                kwargs.get(MessageConstant.GROUP_KEY_CLIENT_ATTRIBUTE_NAME),
                messageCreationRequestKey
            )
            messageCreationRequestHeaders = kwargs.get(MessageConstant.MESSAGE_HEADERS_KEY_CLIENT_ATTRIBUTE_NAME, {})

            completeResponse = [
                MessageDto.MessageCreationRequestDto(
                    key = messageCreationRequestKey,
                    queueKey = resourceMethodQueueKey,
                    groupKey = messageCreationRequestGroupKey
                ),
                {},
                HttpStatus.CREATED
            ]

            emitterArgs = (
                args,
                kwargs,
                wrapperManager,
                requestHeaderClass,
                requestParamClass,
                requestClass,
                responseClass,
                produces,
                httpClientResolversMap,
                returnOnlyBody,
                debugIt,
                messageCreationRequestKey,
                messageCreationRequestGroupKey,
                messageCreationRequestHeaders,
                resourceInstanceMethodMuteStacktraceOnBusinessRuleException
            )

            currentRequestUrl = FlaskUtil.safellyGetUrl()
            if ObjectHelper.isNotNone(currentRequestUrl):
                wrapperManager.resourceInstance.manager.runInAThread(
                    *(
                        resolveEmitterCallWithinAContext,
                        *emitterArgs,
                        currentRequestUrl,
                        FlaskUtil.safellyGetVerb(),
                        FlaskUtil.safellyGetHeaders(),
                        FlaskUtil.safellyGetArgs(),
                        FlaskUtil.safellyGetRequestBody()
                    )
                )
            else:
                log.debug(wrapperManager.resourceInstanceMethod, f'''The context "{currentRequestUrl}" didnt't started properlly. Running without a context by default''')
                wrapperManager.resourceInstance.manager.runInAThread(resolveEmitterCall, *emitterArgs)

            if wrapperManager.shouldLogResponse():
                resourceMethodResponseStatus = completeResponse[-1]
                resourceMethodResponseHeaders = completeResponse[1]
                resourceMethodResponseBody = completeResponse[0] if ObjectHelper.isNotNone(completeResponse[0]) else {'message' : HttpStatus.map(resourceMethodResponseStatus).enumName}
                log.prettyJson(
                    wrapperManager.resourceInstanceMethod,
                    LogConstant.EMITTER_RESPONSE if ObjectHelper.isNotNone(currentRequestUrl) else 'Before Request',
                    {
                        'headers': resourceMethodResponseHeaders,
                        'body': Serializer.getObjectAsDictionary(resourceMethodResponseBody, muteLogs=not debugIt),
                        'status': resourceMethodResponseStatus
                    },
                    condition = True,
                    logLevel = log.INFO
                )
            if returnOnlyBody:
                return completeResponse[0]
            else:
                return completeResponse
        ReflectionHelper.overrideSignatures(innerResourceInstanceMethod, wrapperManager.resourceInstanceMethod)
        innerResourceInstanceMethod.url = resourceMethodConfig.url
        innerResourceInstanceMethod.headers = resourceMethodConfig.headers
        innerResourceInstanceMethod.requestHeaderClass = resourceMethodConfig.requestHeaderClass
        innerResourceInstanceMethod.requestParamClass = resourceMethodConfig.requestParamClass
        innerResourceInstanceMethod.requestClass = resourceMethodConfig.requestClass
        innerResourceInstanceMethod.responseClass = resourceMethodConfig.responseClass
        innerResourceInstanceMethod.returnOnlyBody = resourceMethodConfig.returnOnlyBody
        innerResourceInstanceMethod.timeout = resourceMethodConfig.timeout
        innerResourceInstanceMethod.propagateAuthorization = resourceMethodConfig.propagateAuthorization
        innerResourceInstanceMethod.propagateApiKey = resourceMethodConfig.propagateApiKey
        innerResourceInstanceMethod.propagateSession = resourceMethodConfig.propagateSession
        innerResourceInstanceMethod.produces = resourceMethodConfig.produces
        innerResourceInstanceMethod.consumes = resourceMethodConfig.consumes
        innerResourceInstanceMethod.logRequest = resourceMethodConfig.logRequest
        innerResourceInstanceMethod.logResponse = resourceMethodConfig.logResponse
        innerResourceInstanceMethod.enabled = resourceMethodConfig.enabled
        innerResourceInstanceMethod.muteLogs = resourceMethodConfig.muteLogs
        return innerResourceInstanceMethod
    return innerMethodWrapper


###- https://flask.palletsprojects.com/en/2.1.x/api/#flask.copy_current_request_context
@copy_current_request_context
def resolveEmitterCallWithinAContext(
    args,
    kwargs,
    wrapperManager,
    requestHeaderClass,
    requestParamClass,
    requestClass,
    responseClass,
    produces,
    httpClientResolversMap,
    returnOnlyBody,
    debugIt,
    messageCreationRequestKey,
    messageCreationRequestGroupKey,
    messageCreationRequestHeaders,
    resourceInstanceMethodMuteStacktraceOnBusinessRuleException,

    requestUrl,
    requestVerb,
    requestHeaders,
    requestParams,
    requestBody
):
    ###- https://flask.palletsprojects.com/en/2.1.x/appcontext/
    ###- https://flask.palletsprojects.com/en/2.1.x/reqcontext/
    ###- https://werkzeug.palletsprojects.com/en/2.1.x/test/#werkzeug.test.EnvironBuilder
    # with wrapperManager.api.app.test_request_context(
    #     path = requestUrl,
    #     method = requestVerb,
    #     headers = requestHeaders,
    #     ###- query_string = requestParams, ###- query string already comes in the url
    #     json = requestBody
    # ):
    ###- https://flask.palletsprojects.com/en/1.1.x/appcontext/
    # with wrapperManager.api.app.app_context():
    resolveEmitterCall(
        args,
        kwargs,
        wrapperManager,
        requestHeaderClass,
        requestParamClass,
        requestClass,
        responseClass,
        produces,
        httpClientResolversMap,
        returnOnlyBody,
        debugIt,
        messageCreationRequestKey,
        messageCreationRequestGroupKey,
        messageCreationRequestHeaders,
        resourceInstanceMethodMuteStacktraceOnBusinessRuleException
    )


def resolveEmitterCall(
    args,
    kwargs,
    wrapperManager,
    requestHeaderClass,
    requestParamClass,
    requestClass,
    responseClass,
    produces,
    httpClientResolversMap,
    returnOnlyBody,
    debugIt,
    messageCreationRequestKey,
    messageCreationRequestGroupKey,
    messageCreationRequestHeaders,
    resourceInstanceMethodMuteStacktraceOnBusinessRuleException
):
    resourceMethodResponse = None
    completeResponse = None
    try:
        try:
            FlaskManager.validateKwargs(
                kwargs,
                wrapperManager.resourceInstance,
                wrapperManager.resourceInstanceMethod,
                requestHeaderClass,
                requestParamClass
            )
            FlaskManager.validateArgs(args, requestClass, wrapperManager.resourceInstanceMethod)
            emitterEvent = ClientUtil.getHttpClientEvent(wrapperManager.resourceInstanceMethod, *args, **kwargs)
            if isinstance(emitterEvent, ClientUtil.ManualHttpClientEvent):
                completeResponse = emitterEvent.completeResponse
            elif isinstance(emitterEvent, ClientUtil.HttpClientEvent):
                try:
                    messageKey = messageCreationRequestKey if MessageConstant.MESSAGE_KEY_CLIENT_ATTRIBUTE_NAME not in emitterEvent.kwargs else ConverterStatic.getValueOrDefault(
                        emitterEvent.kwargs.pop(MessageConstant.MESSAGE_KEY_CLIENT_ATTRIBUTE_NAME),
                        messageCreationRequestKey
                    )
                    groupKey = messageCreationRequestGroupKey if MessageConstant.GROUP_KEY_CLIENT_ATTRIBUTE_NAME not in emitterEvent.kwargs else ConverterStatic.getValueOrDefault(
                        emitterEvent.kwargs.pop(MessageConstant.GROUP_KEY_CLIENT_ATTRIBUTE_NAME),
                        messageCreationRequestGroupKey
                    )
                    messageHeaders = messageCreationRequestHeaders if MessageConstant.MESSAGE_HEADERS_KEY_CLIENT_ATTRIBUTE_NAME not in emitterEvent.kwargs else {
                        **ConverterStatic.getValueOrDefault(emitterEvent.kwargs.pop(MessageConstant.MESSAGE_HEADERS_KEY_CLIENT_ATTRIBUTE_NAME), dict()),
                        **ConverterStatic.getValueOrDefault(messageCreationRequestHeaders, dict())
                    }
                    resourceMethodResponse = httpClientResolversMap.get(
                        emitterEvent.verb,
                        ClientUtil.raiseHttpClientEventNotFoundException
                    )(
                        wrapperManager.resourceInstance,
                        *emitterEvent.args,
                        messageKey = messageKey,
                        groupKey = groupKey,
                        messageHeaders = messageHeaders,
                        **emitterEvent.kwargs
                    )
                except Exception as exception:
                    ClientUtil.raiseException(
                        resourceMethodResponse,
                        exception,
                        context = HttpDomain.EMITTER_CONTEXT,
                        businessLogMessage = HttpEmitterConstant.ERROR_AT_CLIENT_CALL_MESSAGE,
                        defaultLogMessage = HttpEmitterConstant.CLIENT_DID_NOT_SENT_ANY_MESSAGE
                    )
                ClientUtil.raiseExceptionIfNeeded(
                    resourceMethodResponse,
                    context = HttpDomain.EMITTER_CONTEXT,
                    businessLogMessage = HttpEmitterConstant.ERROR_AT_CLIENT_CALL_MESSAGE,
                    defaultLogMessage = HttpEmitterConstant.CLIENT_DID_NOT_SENT_ANY_MESSAGE
                )
                completeResponse = ClientUtil.getCompleteResponse(resourceMethodResponse, responseClass, produces)
                FlaskManager.validateCompleteResponse(responseClass, completeResponse)
            else:
                raise Exception('Unknown emitter event')
        except Exception as exception:
            if isinstance(exception, GlobalException):
                log.log(resolveEmitterCall, 'Failure at emitter method execution', exception=exception, muteStackTrace=True)
            else:
                log.failure(resolveEmitterCall, 'Failure at emitter method execution', exception=exception)
            FlaskManager.raiseAndPersistGlobalException(
                exception,
                wrapperManager.resourceInstance,
                wrapperManager.resourceInstanceMethod,
                context = HttpDomain.EMITTER_CONTEXT
            )
    except Exception as exception:
        completeResponse = FlaskManager.getCompleteResponseByException(
            exception,
            wrapperManager.resourceInstance,
            wrapperManager.resourceInstanceMethod,
            resourceInstanceMethodMuteStacktraceOnBusinessRuleException,
            context = HttpDomain.EMITTER_CONTEXT
        )

    if wrapperManager.shouldLogResponse():
        resourceMethodResponseStatus = completeResponse[-1]
        resourceMethodResponseHeaders = completeResponse[1]
        resourceMethodResponseBody = completeResponse[0] if ObjectHelper.isNotNone(completeResponse[0]) else {'message' : HttpStatus.map(resourceMethodResponseStatus).enumName}
        log.prettyJson(
            wrapperManager.resourceInstanceMethod,
            LogConstant.EMITTER_RESPONSE,
            {
                'headers': resourceMethodResponseHeaders,
                'body': Serializer.getObjectAsDictionary(resourceMethodResponseBody, muteLogs=not debugIt),
                'status': resourceMethodResponseStatus
            },
            condition = True,
            logLevel = log.INFO
        )
    if returnOnlyBody:
        return completeResponse[0]
    else:
        return completeResponse



from python_helper import ObjectHelper
from threading import Thread


DEFAULT_TIMEOUT = 20


def runInSimpleThread(target, *args, threadTimeout=DEFAULT_TIMEOUT, **kwargs):
    applicationThread = ApplicationThread(target, *args, threadTimeout=DEFAULT_TIMEOUT, **kwargs)
    applicationThread.run()


class ApplicationThread:

    def __init__(self, target, *args, threadTimeout=DEFAULT_TIMEOUT, **kwargs):
        self.running = False
        self.shouldStop = False
        self.thread = Thread(
            target = target,
            args = args,
            kwargs = kwargs
        )
        self.timeout = threadTimeout


    def run(self, threadTimeout=DEFAULT_TIMEOUT):
        if not self.isRunning():
            self.thread.start()
            self.running = True
            self.shouldStop = False
        # self.thread.join(timeout=threadTimeout if ObjectHelper.isNone(self.timeout) else self.timeout)


    def kill(self):
        self.shouldStop = True
        self.running = False
        del self.thread
        self.thread = None


    def isRunning(self):
        return (self.isAlive() or self.running) and not self.shouldStop


    def isAlive(self):
        return self.thread.is_alive()


    def shouldStopRunning(self):
        return True and self.shouldStop


    def runItIfItsNotRunningYet(self, threadTimeout=DEFAULT_TIMEOUT):
        if not self.isRunning():
            self.run(threadTimeout=threadTimeout)



from python_framework import ConverterStatic

from constant import EmissionConstant


class EmissionRequestDto:
    def __init__(self,
        queueKey = None,
        subscriptionKey = None,
        url = None,
        tries = None,
        onErrorUrl = None,
        onErrorTries = None,
        maxTries = None,
        backOff = None,
        message = None
    ):
        self.queueKey = queueKey
        self.subscriptionKey = subscriptionKey
        self.url = url
        self.tries = ConverterStatic.getValueOrDefault(tries, EmissionConstant.ZERO_TRIES)
        self.onErrorUrl = onErrorUrl
        self.onErrorTries = ConverterStatic.getValueOrDefault(onErrorTries, EmissionConstant.ZERO_TRIES)
        self.maxTries = ConverterStatic.getValueOrDefault(maxTries, EmissionConstant.DEFAULT_MAX_TRIES)
        self.backOff = ConverterStatic.getValueOrDefault(backOff, EmissionConstant.DEFAULT_BACKOFF)
        self.message = message


class EmissionResponseDto:
    def __init__(self,
        queueKey = None,
        subscriptionKey = None,
        url = None,
        tries = None,
        onErrorUrl = None,
        onErrorTries = None,
        maxTries = None,
        backOff = None,
        message = None,
        history = None
    ):
        self.queueKey = queueKey
        self.subscriptionKey = subscriptionKey
        self.url = url
        self.tries = ConverterStatic.getValueOrDefault(tries, EmissionConstant.ZERO_TRIES)
        self.onErrorUrl = onErrorUrl
        self.onErrorTries = ConverterStatic.getValueOrDefault(onErrorTries, EmissionConstant.ZERO_TRIES)
        self.maxTries = ConverterStatic.getValueOrDefault(maxTries, EmissionConstant.DEFAULT_MAX_TRIES)
        self.backOff = ConverterStatic.getValueOrDefault(backOff, EmissionConstant.DEFAULT_BACKOFF)
        self.message = message
        self.history = ConverterStatic.getValueOrDefault(history, [])



from python_framework import Serializer, HttpStatus, HttpClient, HttpClientMethod

from config import MessageConfig
import MessageDto
import Emission


@HttpClient(
    logRequest = True,
    logResponse = True,
    timeout = MessageConfig.EMITTER_TIMEOUT,
    headers = {'Content-Type': 'application/json'}
)
class EmissionClient:

    @HttpClientMethod(
        logRequest = True,
        logResponse = True,
        requestHeaderClass = dict,
        requestClass = [str, MessageDto.MessageRequestDto],
        responseClass = [MessageDto.MessageCreationResponseDto]
    )
    def send(self, url, dto, headers=None):
        return self.post(
            additionalUrl = url,
            headers = headers,
            body = Serializer.getObjectAsDictionary(dto)
        )



from distutils.core import setup
import os

print('''Installation on linux, run:
sudo apt install libpq-dev python3-dev
pip3.9 install --no-cache-dir python-framework --force --upgrade

Aliases:
sudo rm /usr/bin/python
sudo ln -s /usr/local/bin/pythonX.Y /usr/bin/python

sudo rm /usr/bin/pip
sudo ln -s /usr/local/bin/pipX.Y /usr/bin/pip
''')

VERSION = '0.1.26'

SNAKE_CASE_NAME = 'queue_manager_api'
PACKAGE_NAME = SNAKE_CASE_NAME
REPOSITORY_NAME = SNAKE_CASE_NAME.replace('_', '-')
API = 'api'
SRC = 'src'
RESOURCE = 'resource'
URL = f'https://github.com/SamuelJansen/{REPOSITORY_NAME}/'

OS_SEPARATOR = os.path.sep

setup(
    name = SNAKE_CASE_NAME,
    packages = [
        PACKAGE_NAME,
        f'{PACKAGE_NAME}{OS_SEPARATOR}{API}',
        f'{PACKAGE_NAME}{OS_SEPARATOR}{API}{OS_SEPARATOR}{SRC}',
        f'{PACKAGE_NAME}{OS_SEPARATOR}{API}{OS_SEPARATOR}{SRC}{OS_SEPARATOR}annotation',
        f'{PACKAGE_NAME}{OS_SEPARATOR}{API}{OS_SEPARATOR}{SRC}{OS_SEPARATOR}constant',
        f'{PACKAGE_NAME}{OS_SEPARATOR}{API}{OS_SEPARATOR}{SRC}{OS_SEPARATOR}dto',
        f'{PACKAGE_NAME}{OS_SEPARATOR}{API}{OS_SEPARATOR}{SRC}{OS_SEPARATOR}util',
        f'{PACKAGE_NAME}{OS_SEPARATOR}{API}{OS_SEPARATOR}{SRC}{OS_SEPARATOR}manager'
    ],
    # data_files = [
    #     (STATIC_PACKAGE_PATH, [
    #         f'{RELATIVE_PATH}{OS_SEPARATOR}resource_1.extension',
    #         f'{RELATIVE_PATH}{OS_SEPARATOR}resource_2.extension'
    #     ])
    # ],
    version = VERSION,
    license = 'MIT',
    description = 'Queue Manager',
    author = 'Samuel Jansen',
    author_email = 'samuel.jansenn@gmail.com',
    url = URL,
    download_url = f'{URL}archive/v{VERSION}.tar.gz',
    keywords = ['queue', 'topic'],
    install_requires = [
        'python-framework<1.0.0,>=0.3.91',
        'globals<1.0,>=0.3.34',
        'python-helper<1.0,>=0.3.51'
    ],
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7'
    ],
    python_requires='>=3.7'
)



from python_framework import Enum, EnumItem


@Enum()
class ModelStatusEnumeration :
    ACCEPTED = EnumItem()
    PROCESSING = EnumItem()
    PROCESSED = EnumItem()
    PROCESSED_WITH_ERRORS = EnumItem()
    UNPROCESSED = EnumItem()
    DELIVERED = EnumItem()
    ERROR = EnumItem()
    NOT_INFORMED = EnumItem()


ModelStatus = ModelStatusEnumeration()



from python_helper import log, ObjectHelper
from python_framework import Service, ServiceMethod

import Emission, EmissionModel
from constant import EmissionConstant
from enumeration.ModelState import ModelState
from enumeration.ModelStatus import ModelStatus


LOG_LEVEL = log.STATUS


@Service()
class EmissionModelService:

    @ServiceMethod(requestClass=[[Emission.Emission]])
    def createOrUpdateAll(self, emissionList):
        if ObjectHelper.isEmpty(emissionList):
            return []
        self.mapper.emission.overrideAllModelState(emissionList, ModelState.PERSISTED)
        modelList = self.mapper.emissionModel.fromRequestDtoListToModelList(emissionList)
        log.prettyPython(self.createOrUpdateAll, f'Creating or updating queued message emissions', modelList, logLevel=LOG_LEVEL)
        existingModelList = self.findAllModelByKeyIn(list({model.key for model in modelList}))
        existingModelDictionary = {
            self.helper.emissionModel.getModelKey(model): model for model in existingModelList
        }
        for model in modelList:
            if self.helper.emissionModel.getModelKey(model) in existingModelDictionary:
                self.mapper.emissionModel.overrideModel(
                    existingModelDictionary.get(self.helper.emissionModel.getModelKey(model)),
                    model
                )
            else:
                existingModelList.append(model)
        return self.persistAll(existingModelList)


    @ServiceMethod(requestClass=[[str]])
    def findAllModelByKeyIn(self, keyList):
        return self.repository.emissionModel.findAllByKeyIn(keyList)


    @ServiceMethod(requestClass=[str])
    def existsByKey(self, key):
        return self.repository.emissionModel.existsByKey(key)


    @ServiceMethod(requestClass=[[EmissionModel.EmissionModel]])
    def persistAll(self, modelList):
        self.mapper.messageModel.overrideAllModelState(modelList, ModelState.PERSISTED)
        log.prettyPython(self.createOrUpdateAll, f'Persisting queued message emissions', modelList, logLevel=LOG_LEVEL)
        return self.repository.messageModel.saveAll(modelList)



from python_framework import SchedulerType
from python_framework import Scheduler, SchedulerMethod, WeekDay, WeekDayConstant


@Scheduler(muteLogs=True)
class MessageScheduler:


    @SchedulerMethod(SchedulerType.INTERVAL, seconds=1, instancesUpTo=5)
    def updateAllModifiedFromMemory(self) :
        self.service.message.updateAllModifiedFromMemory()


    @SchedulerMethod(SchedulerType.INTERVAL, seconds=0.1, instancesUpTo=100)
    def sendAllAcceptedFromOneQueue(self) :
        self.service.message.sendAllAcceptedFromOneQueue()



class MessageRequestDto:
    def __init__(self,
        key = None,
        queueKey = None,
        groupKey = None,
        headers = None,
        content = None
    ):
        self.key = key
        self.queueKey = queueKey
        self.groupKey = groupKey
        self.headers = headers
        self.content = content


class MessageResponseDto:
    def __init__(self,
        key = None,
        queueKey = None,
        groupKey = None,
        content = None,
        history = None
    ):
        self.key = key
        self.queueKey = queueKey
        self.groupKey = groupKey
        self.content = content
        self.history = history

class MessageCreationRequestDto:
    def __init__(self,
        key = None,
        queueKey = None,
        groupKey = None
    ):
        self.key = key
        self.queueKey = queueKey
        self.groupKey = groupKey

class MessageCreationResponseDto:
    def __init__(self,
        key = None,
        queueKey = None,
        groupKey = None
    ):
        self.key = key
        self.queueKey = queueKey
        self.groupKey = groupKey
